/* bnx2i_iscsi.c: Broadcom NetXtreme II iSCSI driver.
 *
 * Copyright (c) 2006 - 2009 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: Anil Veerabhadrappa (anilgv@broadcom.com)
 */

#include "bnx2i.h"
#include <linux/ethtool.h>
#include <scsi/scsi_transport_iscsi.h>

struct scsi_host_template bnx2i_host_template;
struct iscsi_transport bnx2i_iscsi_transport;
struct file_operations bnx2i_mgmt_fops;
extern unsigned int bnx2i_nopout_when_cmds_active;

/*
 * Global endpoint resource info
 */
static void *bnx2i_ep_pages[MAX_PAGES_PER_CTRL_STRUCT_POOL];
static struct list_head bnx2i_free_ep_list;
static struct list_head bnx2i_unbound_ep;
static u32 bnx2i_num_free_ep;
static u32 bnx2i_max_free_ep;
static DEFINE_SPINLOCK(bnx2i_resc_lock); /* protects global resources */
struct tcp_port_mngt bnx2i_tcp_port_tbl;

extern unsigned int sq_size;
extern unsigned int rq_size;

/* Char device major number */
static int bnx2i_major_no;

static struct io_bdt *bnx2i_alloc_bd_table(struct bnx2i_sess *sess,
					   struct bnx2i_cmd *);

void bnx2i_unbind_adapter_devices(struct bnx2i_hba *hba);

/*
 * iSCSI Session's hostdata organization:
 *
 *    *------------------* <== hostdata_session(host->hostdata)
 *    | ptr to class sess|
 *    |------------------| <== iscsi_hostdata(host->hostdata)
 *    | iscsi_session    |
 *    *------------------*
 */

#define hostdata_privsize(_sz)	(sizeof(unsigned long) + _sz + \
				 _sz % sizeof(unsigned long))
#define hostdata_session(_hostdata) (iscsi_ptr(*(unsigned long *)_hostdata))

#define session_to_cls(_sess) 	hostdata_session(_sess->host->hostdata)

/**
 * bnx2i_alloc_tcp_port - allocates a tcp port from the free list
 *
 * assumes this function is called with 'bnx2i_resc_lock' held
 **/
static u16 bnx2i_alloc_tcp_port(void)
{
	u16 tcp_port;

	if (!bnx2i_tcp_port_tbl.num_free_ports || !bnx2i_tcp_port_tbl.free_q)
		return 0;

	tcp_port = bnx2i_tcp_port_tbl.free_q[bnx2i_tcp_port_tbl.cons_idx];
	bnx2i_tcp_port_tbl.cons_idx++;
	bnx2i_tcp_port_tbl.cons_idx %= bnx2i_tcp_port_tbl.max_idx;
	bnx2i_tcp_port_tbl.num_free_ports--;

	return tcp_port;
}


/**
 * bnx2i_free_tcp_port - Frees the given tcp port back to free pool
 *
 * @port: 		tcp port number being freed
 *
 * assumes this function is called with 'bnx2i_resc_lock' held
 **/
static void bnx2i_free_tcp_port(u16 port)
{
	if (!bnx2i_tcp_port_tbl.free_q)
		return;

	bnx2i_tcp_port_tbl.free_q[bnx2i_tcp_port_tbl.prod_idx] = port;
	bnx2i_tcp_port_tbl.prod_idx++;
	bnx2i_tcp_port_tbl.prod_idx %= bnx2i_tcp_port_tbl.max_idx;
	bnx2i_tcp_port_tbl.num_free_ports++;
}

/**
 * bnx2i_tcp_port_new_entry - place 'bnx2id' allocated tcp port number
 *		to free list
 *
 * @port: 		tcp port number being added to free pool
 *
 * 'bnx2i_resc_lock' is held while operating on global tcp port table
 **/
void bnx2i_tcp_port_new_entry(u16 tcp_port)
{
	u32 idx = bnx2i_tcp_port_tbl.prod_idx;

	spin_lock(&bnx2i_resc_lock);
	bnx2i_tcp_port_tbl.free_q[idx] = tcp_port;
	bnx2i_tcp_port_tbl.prod_idx++;
	bnx2i_tcp_port_tbl.prod_idx %= bnx2i_tcp_port_tbl.max_idx;
	bnx2i_tcp_port_tbl.num_free_ports++;
	bnx2i_tcp_port_tbl.num_required--;
	spin_unlock(&bnx2i_resc_lock);
}

/**
 * bnx2i_init_tcp_port_mngr - initializes tcp port manager
 *
 */
void bnx2i_init_tcp_port_mngr(void)
{
	int mem_size;

	bnx2i_tcp_port_tbl.num_free_ports = 0;
	bnx2i_tcp_port_tbl.prod_idx = 0;
	bnx2i_tcp_port_tbl.cons_idx = 0;
	bnx2i_tcp_port_tbl.max_idx = 0;
	bnx2i_tcp_port_tbl.num_required = 0;

#define BNX2I_MAX_TCP_PORTS	1024

	bnx2i_tcp_port_tbl.port_tbl_size = BNX2I_MAX_TCP_PORTS;

	mem_size = sizeof(u16) * bnx2i_tcp_port_tbl.port_tbl_size;
	if (bnx2i_tcp_port_tbl.port_tbl_size) {
		bnx2i_tcp_port_tbl.free_q = kmalloc(mem_size, GFP_KERNEL);

		if (bnx2i_tcp_port_tbl.free_q)
			bnx2i_tcp_port_tbl.max_idx =
				bnx2i_tcp_port_tbl.port_tbl_size;
	}
}


/**
 * bnx2i_cleanup_tcp_port_mngr - frees memory held by global tcp port table
 *
 */
void bnx2i_cleanup_tcp_port_mngr(void)
{
	kfree(bnx2i_tcp_port_tbl.free_q);
	bnx2i_tcp_port_tbl.free_q = NULL;
	bnx2i_tcp_port_tbl.num_free_ports = 0;
}

static int bnx2i_adapter_ready(struct bnx2i_hba *hba)
{
	int retval = 0;

	if (!hba || !test_bit(ADAPTER_STATE_UP, &hba->adapter_state) ||
	    test_bit(ADAPTER_STATE_GOING_DOWN, &hba->adapter_state) ||
	    test_bit(ADAPTER_STATE_LINK_DOWN, &hba->adapter_state))
		retval = -EPERM;
	return retval;
}

/**
 * bnx2i_get_write_cmd_bd_idx - identifies various BD bookmarks for a
 *			scsi write command
 *
 * @cmd:		iscsi cmd struct pointer
 * @buf_off:		absolute buffer offset
 * @start_bd_off:	u32 pointer to return the offset within the BD
 *			indicated by 'start_bd_idx' on which 'buf_off' falls
 * @start_bd_idx:	index of the BD on which 'buf_off' falls
 *
 * identifies & marks various bd info for imm data, unsolicited data
 *	and the first solicited data seq.
 */
static void bnx2i_get_write_cmd_bd_idx(struct bnx2i_cmd *cmd, u32 buf_off,
				       u32 *start_bd_off, u32 *start_bd_idx)
{
	u32 cur_offset = 0;
	u32 cur_bd_idx = 0;
	struct iscsi_bd *bd_tbl = cmd->bd_tbl->bd_tbl;

	if (buf_off) {
		while (buf_off >= (cur_offset + bd_tbl->buffer_length)) {
			cur_offset += bd_tbl->buffer_length;
			cur_bd_idx++;
			bd_tbl++;
		}
	}

	*start_bd_off = buf_off - cur_offset;
	*start_bd_idx = cur_bd_idx;
}

/**
 * bnx2i_setup_write_cmd_bd_info - sets up BD various information for
 *			scsi write command
 *
 * @cmd:		iscsi cmd struct pointer
 *
 * identifies & marks various bd info for immediate data, unsolicited data
 *	and first solicited data seq which includes BD start index & BD buf off
 *	This function takes into account iscsi parameter such as immediate data
 *	and unsolicited data is support on this connection
 *	
 */
static void bnx2i_setup_write_cmd_bd_info(struct bnx2i_cmd *cmd)
{
	struct bnx2i_sess *sess;
	u32 start_bd_offset;
	u32 start_bd_idx; 
	u32 buffer_offset = 0;
	u32 seq_len = 0;
	u32 fbl, mrdsl;
	u32 cmd_len = cmd->req.total_data_transfer_length;

	sess = cmd->conn->sess;

	/* if ImmediateData is turned off & IntialR2T is turned on,
	 * there will be no immediate or unsolicited data, just return.
	 */
	if (sess->initial_r2t && !sess->imm_data)
		return;

	fbl = sess->first_burst_len;
	mrdsl = cmd->conn->max_data_seg_len_xmit;

	/* Immediate data */
	if (sess->imm_data) {
		seq_len = min(mrdsl, fbl);
		seq_len = min(cmd_len, seq_len);
		buffer_offset += seq_len;
	}

	if (seq_len == cmd_len)
		return;

	if (!sess->initial_r2t) {
		if (seq_len >= fbl)
			goto r2t_data;
		seq_len = min(fbl, cmd_len) - seq_len;
		bnx2i_get_write_cmd_bd_idx(cmd, buffer_offset,
					   &start_bd_offset, &start_bd_idx);
		cmd->req.ud_buffer_offset = start_bd_offset;
		cmd->req.ud_start_bd_index = start_bd_idx;
		buffer_offset += seq_len;
	}
r2t_data:
	if (buffer_offset != cmd_len) {
		bnx2i_get_write_cmd_bd_idx(cmd, buffer_offset,
					   &start_bd_offset, &start_bd_idx);
		if ((start_bd_offset > fbl) ||
		    (start_bd_idx > scsi_sg_count(cmd->scsi_cmd))) {
			int i = 0;

			printk(KERN_ALERT "bnx2i- error, buf offset 0x%x "
					  "bd_valid %d use_sg %d\n",
					  buffer_offset, cmd->bd_tbl->bd_valid,
					  scsi_sg_count(cmd->scsi_cmd));
			for (i = 0; i < cmd->bd_tbl->bd_valid; i++)
				printk(KERN_ALERT "bnx2i err, bd[%d]: len %x\n",
						  i, cmd->bd_tbl->bd_tbl[i].\
						  buffer_length);
		}
		cmd->req.sd_buffer_offset = start_bd_offset;
		cmd->req.sd_start_bd_index = start_bd_idx;
	}
}


/**
 * bnx2i_split_bd - splits buffer > 64KB into 32KB chunks
 *
 * @cmd:		iscsi cmd struct pointer
 * @addr: 		base address of the buffer
 * @sg_len: 		buffer length
 * @bd_index: 		starting index into BD table
 *
 * This is not required as driver limits max buffer size of less than 64K by
 *	advertising 'max_sectors' within this limit. 5706/5708 hardware limits
 *	BD length to less than or equal to 0xFFFF 
 **/
static int bnx2i_split_bd(struct bnx2i_cmd *cmd, u64 addr, int sg_len,
			  int bd_index)
{
	struct iscsi_bd *bd = cmd->bd_tbl->bd_tbl;
	int frag_size, sg_frags;

	sg_frags = 0;
	while (sg_len) {
		if (sg_len >= BD_SPLIT_SIZE)
			frag_size = BD_SPLIT_SIZE;
		else
			frag_size = sg_len;
		bd[bd_index + sg_frags].buffer_addr_lo = (u32) addr;
		bd[bd_index + sg_frags].buffer_addr_hi = addr >> 32;
		bd[bd_index + sg_frags].buffer_length = frag_size;
		bd[bd_index + sg_frags].flags = 0;
		if ((bd_index + sg_frags) == 0)
			bd[0].flags = ISCSI_BD_FIRST_IN_BD_CHAIN;
		addr += (u64) frag_size;
		sg_frags++;
		sg_len -= frag_size;
	}
	return sg_frags;
}


/**
 * bnx2i_map_single_buf - maps a single buffer and updates the BD table
 *
 * @hba: 		adapter instance
 * @cmd:		iscsi cmd struct pointer
 *
 */
static int bnx2i_map_single_buf(struct bnx2i_hba *hba,
				       struct bnx2i_cmd *cmd)
{
	struct scsi_cmnd *sc = cmd->scsi_cmd;
	struct iscsi_bd *bd = cmd->bd_tbl->bd_tbl;
	int byte_count;
	int bd_count;
	u64 addr;

	byte_count = scsi_bufflen(sc);
	sc->SCp.dma_handle =
		pci_map_single(hba->pcidev, scsi_sglist(sc),
			       scsi_bufflen(sc), sc->sc_data_direction);
	addr = sc->SCp.dma_handle;

	if (byte_count > MAX_BD_LENGTH) {
		bd_count = bnx2i_split_bd(cmd, addr, byte_count, 0);
	} else {
		bd_count = 1;
		bd[0].buffer_addr_lo = addr & 0xffffffff;
		bd[0].buffer_addr_hi = addr >> 32;
		bd[0].buffer_length = scsi_bufflen(sc);
		bd[0].flags = ISCSI_BD_FIRST_IN_BD_CHAIN |
			      ISCSI_BD_LAST_IN_BD_CHAIN;
	}
	bd[bd_count - 1].flags |= ISCSI_BD_LAST_IN_BD_CHAIN;

	return bd_count;
}


/**
 * bnx2i_map_sg - maps IO buffer and prepares the BD table
 *
 * @hba: 		adapter instance
 * @cmd:		iscsi cmd struct pointer
 *
 * map SG list
 */
static int bnx2i_map_sg(struct bnx2i_hba *hba, struct bnx2i_cmd *cmd)
{
	struct scsi_cmnd *sc = cmd->scsi_cmd;
	struct iscsi_bd *bd = cmd->bd_tbl->bd_tbl;
	struct scatterlist *sg;
	int byte_count = 0;
	int sg_frags;
	int bd_count = 0;
	int sg_count;
	int sg_len;
	u64 addr;
	int i;

	sg = scsi_sglist(sc);
	sg_count = pci_map_sg(hba->pcidev, sg, scsi_sg_count(sc),
			      sc->sc_data_direction);

	for (i = 0; i < sg_count; i++) {
		sg_len = sg_dma_len(sg);
		addr = sg_dma_address(sg);
		if (sg_len > MAX_BD_LENGTH)
			sg_frags = bnx2i_split_bd(cmd, addr, sg_len,
						  bd_count);
		else {
			sg_frags = 1;
			bd[bd_count].buffer_addr_lo = addr & 0xffffffff;
			bd[bd_count].buffer_addr_hi = addr >> 32;
			bd[bd_count].buffer_length = sg_len;
			bd[bd_count].flags = 0;
			if (bd_count == 0)
				bd[bd_count].flags =
					ISCSI_BD_FIRST_IN_BD_CHAIN;
		}
		byte_count += sg_len;
		sg++;
		bd_count += sg_frags;
	}
	bd[bd_count - 1].flags |= ISCSI_BD_LAST_IN_BD_CHAIN;

	BUG_ON(byte_count != scsi_bufflen(sc));
	return bd_count;
}

/**
 * bnx2i_iscsi_map_sg_list - maps SG list
 *
 * @cmd:		iscsi cmd struct pointer
 *
 * creates BD list table for the command
 */
static void bnx2i_iscsi_map_sg_list(struct bnx2i_cmd *cmd)
{
	struct scsi_cmnd *sc = cmd->scsi_cmd;
	int bd_count;

	if (scsi_sg_count(sc))
		bd_count = bnx2i_map_sg(cmd->conn->sess->hba, cmd);
	else if (scsi_bufflen(sc))
		bd_count = bnx2i_map_single_buf(cmd->conn->sess->hba, cmd);
	else {
		struct iscsi_bd *bd = cmd->bd_tbl->bd_tbl;
		bd_count  = 0;
		bd[0].buffer_addr_lo = bd[0].buffer_addr_hi = 0;
		bd[0].buffer_length = bd[0].flags = 0;
	}
	cmd->bd_tbl->bd_valid = bd_count;
}


/**
 * bnx2i_iscsi_unmap_sg_list - unmaps SG list
 *
 * @cmd:		iscsi cmd struct pointer
 *
 * unmap IO buffers and invalidate the BD table
 */
void bnx2i_iscsi_unmap_sg_list(struct bnx2i_hba *hba, struct bnx2i_cmd *cmd)
{
	struct scsi_cmnd *sc = cmd->scsi_cmd;
	struct scatterlist *sg;

	if (cmd->bd_tbl->bd_valid && sc) {
		if (scsi_sg_count(sc)) {
			sg = scsi_sglist(sc);
			pci_unmap_sg(hba->pcidev, sg, scsi_sg_count(sc),
				     sc->sc_data_direction);
		} else {
			pci_unmap_single(hba->pcidev, sc->SCp.dma_handle,
					 scsi_bufflen(sc),
					 sc->sc_data_direction);
		}
		cmd->bd_tbl->bd_valid = 0;
	}
}



static void bnx2i_setup_cmd_wqe_template(struct bnx2i_cmd *cmd)
{
	memset(&cmd->req, 0x00, sizeof(cmd->req));
	cmd->req.op_code = 0xFF;
	cmd->req.bd_list_addr_lo = (u32) cmd->bd_tbl->bd_tbl_dma;
	cmd->req.bd_list_addr_hi =
		(u32) ((u64) cmd->bd_tbl->bd_tbl_dma >> 32);

}


/**
 * bnx2i_bind_conn_to_iscsi_cid - bind conn structure to 'iscsi_cid'
 *
 * @conn: 		pointer to iscsi connection
 * @iscsi_cid:		iscsi context ID, range 0 - (MAX_CONN - 1)
 *
 * update iscsi cid table entry with connection pointer. This enables
 *	driver to quickly get hold of connection structure pointer in
 *	completion/interrupt thread using iscsi context ID
 */
static int bnx2i_bind_conn_to_iscsi_cid(struct bnx2i_conn *conn,
					 u32 iscsi_cid)
{
	struct bnx2i_hba *hba;

	if (!conn || !conn->sess)
		return -EINVAL;

	hba = conn->sess->hba;

	if (hba && hba->cid_que.conn_cid_tbl[iscsi_cid]) {
		printk(KERN_ERR "bnx2i: conn bind - entry #%d not free\n",
				iscsi_cid);
		return -EBUSY;
	}

	hba->cid_que.conn_cid_tbl[iscsi_cid] = conn;
	return 0;
}


/**
 * bnx2i_get_conn_from_id - maps an iscsi cid to corresponding conn ptr
 * 
 * @hba: 		pointer to adapter instance
 * @iscsi_cid:		iscsi context ID, range 0 - (MAX_CONN - 1)
 */
struct bnx2i_conn *bnx2i_get_conn_from_id(struct bnx2i_hba *hba,
						 u16 iscsi_cid)
{
	if (!hba->cid_que.conn_cid_tbl) {
		printk(KERN_ERR "bnx2i: ERROR - missing conn<->cid table\n");
		return NULL;

	} else if (iscsi_cid >= hba->max_active_conns) {
		printk(KERN_ERR "bnx2i: wrong cid #%d\n", iscsi_cid);
		return NULL;
	}
	return hba->cid_que.conn_cid_tbl[iscsi_cid];
}


/**
 * bnx2i_alloc_iscsi_cid - allocates a iscsi_cid from free pool
 *
 * @hba: 		pointer to adapter instance
 */
static u32 bnx2i_alloc_iscsi_cid(struct bnx2i_hba *hba)
{
	int idx;

	if (!hba->cid_que.cid_free_cnt)
		return ISCSI_RESERVED_TAG;

	idx = hba->cid_que.cid_q_cons_idx;
	hba->cid_que.cid_q_cons_idx++;
	if (hba->cid_que.cid_q_cons_idx == hba->cid_que.cid_q_max_idx)
		hba->cid_que.cid_q_cons_idx = 0;

	hba->cid_que.cid_free_cnt--;
	return hba->cid_que.cid_que[idx];
}


/**
 * bnx2i_free_iscsi_cid - returns tcp port to free list
 *
 * @hba: 		pointer to adapter instance
 * @iscsi_cid:		iscsi context ID to free
 */
static void bnx2i_free_iscsi_cid(struct bnx2i_hba *hba, u16 iscsi_cid)
{
	int idx;

	if (iscsi_cid == (u16)ISCSI_RESERVED_TAG)
		return;

	hba->cid_que.cid_free_cnt++;

	idx = hba->cid_que.cid_q_prod_idx;
	hba->cid_que.cid_que[idx] = iscsi_cid;
	hba->cid_que.conn_cid_tbl[iscsi_cid] = NULL;
	hba->cid_que.cid_q_prod_idx++;
	if (hba->cid_que.cid_q_prod_idx == hba->cid_que.cid_q_max_idx)
		hba->cid_que.cid_q_prod_idx = 0;
}


/**
 * bnx2i_setup_free_cid_que - sets up free iscsi cid queue
 *
 * @hba: 		pointer to adapter instance
 *
 * allocates memory for iscsi cid queue & 'cid - conn ptr' mapping table,
 * 	and initialize table attributes
 */
static int bnx2i_setup_free_cid_que(struct bnx2i_hba *hba)
{
	int mem_size;
	int i;

	mem_size = hba->max_active_conns * sizeof(u32);
	mem_size = (mem_size + (PAGE_SIZE - 1)) & PAGE_MASK;

	hba->cid_que.cid_que_base = kmalloc(mem_size, GFP_KERNEL);
	if (!hba->cid_que.cid_que_base)
		return -ENOMEM;

	mem_size = hba->max_active_conns * sizeof(struct bnx2i_conn *);
	mem_size = (mem_size + (PAGE_SIZE - 1)) & PAGE_MASK;
	hba->cid_que.conn_cid_tbl = kmalloc(mem_size, GFP_KERNEL);
	if (!hba->cid_que.conn_cid_tbl) {
		kfree(hba->cid_que.cid_que_base);
		hba->cid_que.cid_que_base = NULL;
		return -ENOMEM;
	}

	hba->cid_que.cid_que = (u32 *)hba->cid_que.cid_que_base;
	hba->cid_que.cid_q_prod_idx = 0;
	hba->cid_que.cid_q_cons_idx = 0;
	hba->cid_que.cid_q_max_idx = hba->max_active_conns;
	hba->cid_que.cid_free_cnt = hba->max_active_conns;

	for (i = 0; i < hba->max_active_conns; i++) {
		hba->cid_que.cid_que[i] = i;
		hba->cid_que.conn_cid_tbl[i] = NULL;
	}
	return 0;
}


/**
 * bnx2i_release_free_cid_que - releases 'iscsi_cid' queue resources
 *
 * @hba: 		pointer to adapter instance
 */
static void bnx2i_release_free_cid_que(struct bnx2i_hba *hba)
{
	kfree(hba->cid_que.cid_que_base);
	hba->cid_que.cid_que_base = NULL;

	kfree(hba->cid_que.conn_cid_tbl);
	hba->cid_que.conn_cid_tbl = NULL;
}


/**
 * bnx2i_alloc_ep - allocates ep structure from global pool
 *
 * @hba: 		pointer to adapter instance
 *
 * routine allocates a free endpoint structure from global pool and
 *	a tcp port to be used for this connection.  Global resource lock,
 *	'bnx2i_resc_lock' is held while accessing shared global data structures
 */
static struct bnx2i_endpoint *bnx2i_alloc_ep(struct bnx2i_hba *hba)
{
	struct bnx2i_endpoint *endpoint;
	struct list_head *listp;
	u16 tcp_port;
	u32 flags;

	spin_lock_irqsave(&bnx2i_resc_lock, flags);

	tcp_port = bnx2i_alloc_tcp_port();
	if (!tcp_port) {
		printk(KERN_ERR "bnx2i: unable to allocate tcp ports, "
				"make sure 'bnx2id' is running\n");
		spin_unlock_irqrestore(&bnx2i_resc_lock, flags);
		return NULL;
	}
	if (list_empty(&bnx2i_free_ep_list)) {
		spin_unlock_irqrestore(&bnx2i_resc_lock, flags);
		printk(KERN_ERR "bnx2i: ep struct pool empty\n");
		return NULL;
	}
	listp = (struct list_head *) bnx2i_free_ep_list.next;
	list_del_init(listp);
	bnx2i_num_free_ep--;

	endpoint = (struct bnx2i_endpoint *) listp;
	endpoint->state = EP_STATE_IDLE;
	endpoint->hba = hba;
	endpoint->hba_age = hba->age;
	hba->ofld_conns_active++;
	endpoint->tcp_port = tcp_port;
	init_waitqueue_head(&endpoint->ofld_wait);

	spin_unlock_irqrestore(&bnx2i_resc_lock, flags);
	return endpoint;
}


/**
 * bnx2i_free_ep - returns endpoint struct and tcp port to free pool
 *
 * @endpoint:		pointer to endpoint structure
 *
 */
static void bnx2i_free_ep(struct bnx2i_endpoint *endpoint)
{
	u32 flags;

	spin_lock_irqsave(&bnx2i_resc_lock, flags);
	endpoint->state = EP_STATE_IDLE;
	endpoint->hba->ofld_conns_active--;

	bnx2i_free_iscsi_cid(endpoint->hba, endpoint->ep_iscsi_cid);
	if (endpoint->conn) {
		endpoint->conn->ep = NULL;
		endpoint->conn = NULL;
	}
	endpoint->sess = NULL;

	if (endpoint->tcp_port)
		bnx2i_free_tcp_port(endpoint->tcp_port);

	endpoint->hba = NULL;
	list_add_tail(&endpoint->link, &bnx2i_free_ep_list);
	bnx2i_num_free_ep++;
	spin_unlock_irqrestore(&bnx2i_resc_lock, flags);
}


/**
 * bnx2i_alloc_ep_pool - alloccates a pool of endpoint structures
 *
 * allocates free pool of endpoint structures, which is used to store
 *	QP related control & PT info and other option-2 information
 */
int bnx2i_alloc_ep_pool(void)
{
	struct bnx2i_endpoint *endpoint;
	int index;
	int ret_val = 1;
	int total_endpoints;
	int page_count = 0;
	void *mem_ptr;
	int mem_size;

	spin_lock_init(&bnx2i_resc_lock);
	INIT_LIST_HEAD(&bnx2i_free_ep_list);
	INIT_LIST_HEAD(&bnx2i_unbound_ep);

	for (index = 0; index < MAX_PAGES_PER_CTRL_STRUCT_POOL; index++) {
		bnx2i_ep_pages[index] = NULL;
	}

	total_endpoints = ISCSI_MAX_CONNS_PER_HBA * ISCSI_MAX_ADAPTERS;
	bnx2i_num_free_ep = 0;
	mem_size = total_endpoints * sizeof(struct bnx2i_endpoint);
	mem_ptr = vmalloc(mem_size);
	if (!mem_ptr) {
		printk(KERN_ERR "ep_pool: mem alloc failed\n");
		goto mem_alloc_err;
	}
	memset(mem_ptr, 0, mem_size);

	bnx2i_ep_pages[page_count++] = mem_ptr;
	endpoint = mem_ptr;

	for (index = 0; index < total_endpoints; index++) {
		INIT_LIST_HEAD(&endpoint->link);
		list_add_tail(&endpoint->link, &bnx2i_free_ep_list);
		endpoint++;
		bnx2i_num_free_ep++;
	}
mem_alloc_err:
	if (bnx2i_num_free_ep == 0)
		ret_val = 0;

	bnx2i_max_free_ep = bnx2i_num_free_ep;
	return ret_val;
}


/**
 * bnx2i_release_ep_pool - releases memory resources held by endpoint structs
 */
void bnx2i_release_ep_pool(void)
{
	int index;
	void *mem_ptr;

	for (index = 0; index < MAX_PAGES_PER_CTRL_STRUCT_POOL; index++) {
		mem_ptr = bnx2i_ep_pages[index];
		vfree(mem_ptr);
		bnx2i_ep_pages[index] = NULL;
	}
	bnx2i_num_free_ep = 0;
	return;
}


/**
 * bnx2i_alloc_itt - allocates an ITT from the free pool
 *
 * @sess:		iscsi session pointer
 * @cmd:		iscsi cmd pointer
 *
 * iSCSI Session ITT queue manager
 */
static u32 bnx2i_alloc_itt(struct bnx2i_sess *sess, struct bnx2i_cmd *cmd)
{
	u32 itt_val = ITT_INVALID_SIGNATURE;

	if (sess->itt_q.itt_q_count) {
		itt_val = sess->itt_q.itt_que[sess->itt_q.itt_q_cons_idx++];
		sess->itt_q.itt_q_cons_idx %= sess->itt_q.itt_q_max_idx;
		sess->itt_q.itt_cmd[itt_val] = cmd;
		sess->itt_q.itt_q_count--;
	}
	return itt_val & ISCSI_CMD_REQUEST_INDEX;
}


/**
 * bnx2i_free_itt - releases ITT back to free pool
 *
 * @sess:		iscsi session pointer
 * @cmd:		iscsi cmd pointer
 *
 * iSCSI Session ITT queue manager
 */
static void bnx2i_free_itt(struct bnx2i_sess *sess, struct bnx2i_cmd *cmd)
{
/* check, BUG_ON() */
	if (cmd->req.itt == ITT_INVALID_SIGNATURE) {
		printk(KERN_ALERT "free_itt: RSVD ITT - sess 0x%p\n", sess);
		return;
	}
	sess->itt_q.itt_que[sess->itt_q.itt_q_prod_idx++] =
		cmd->req.itt & ISCSI_CMD_RESPONSE_INDEX;
	sess->itt_q.itt_q_prod_idx %= sess->itt_q.itt_q_max_idx;
	sess->itt_q.itt_cmd[cmd->req.itt & ISCSI_CMD_RESPONSE_INDEX] = NULL;
	sess->itt_q.itt_q_count++;
	cmd->req.itt = ITT_INVALID_SIGNATURE;
}


/**
 * bnx2i_setup_free_itt_queue - allocates a tcp port from the free list
 *
 * @sess:		iscsi session pointer
 *
 * Allocate memory for ITT queue area and setup queue atrributes. ITT queue
 *	is a circular array of ITTs [range 0 - (SQ SIZE-1)] managed by
 *	producer / consumer indexes
 */
static int bnx2i_setup_free_itt_queue(struct bnx2i_sess *sess)
{
	u16 itt_q_size = (u16)sess->sq_size;
	u32 itt_value;
	int unit_size = sizeof(u16);
	int mem_size = PAGE_SIZE;

	if ((itt_q_size * unit_size) > mem_size)
		mem_size = (itt_q_size * unit_size);

	mem_size = (mem_size + (PAGE_SIZE - 1)) & PAGE_MASK;
	sess->itt_q.itt_que_base = kmalloc(mem_size, GFP_KERNEL);
	if (!sess->itt_q.itt_que_base)
		return -ENOMEM;

	mem_size = (itt_q_size * sizeof(struct bnx2i_cmd *));
	mem_size = (mem_size + (PAGE_SIZE - 1)) & PAGE_MASK;
	sess->itt_q.itt_cmd = kmalloc(mem_size, GFP_KERNEL);
	if (!sess->itt_q.itt_cmd) {
		kfree(sess->itt_q.itt_que_base);
		sess->itt_q.itt_que_base = NULL;
		return -1;
	}
	memset(sess->itt_q.itt_cmd, 0x00, mem_size);

	sess->itt_q.itt_que = (u32 *)sess->itt_q.itt_que_base;
	sess->itt_q.itt_q_prod_idx = 0;
	sess->itt_q.itt_q_cons_idx = 0;
	sess->itt_q.itt_q_max_idx = itt_q_size;
	sess->itt_q.itt_q_count = itt_q_size;

	itt_value = 0;
	while (itt_value < itt_q_size) {
		sess->itt_q.itt_cmd[itt_value] = NULL;
		sess->itt_q.itt_que[sess->itt_q.itt_q_prod_idx++] =
			itt_value++;
		if (sess->itt_q.itt_q_prod_idx >= sess->itt_q.itt_q_max_idx)
			sess->itt_q.itt_q_prod_idx = 0;
	}

	return 0;
}


/**
 * bnx2i_release_free_itt_queue - releases resources held by the ITT queue
 *
 * @sess:		iscsi session pointer
 *
 * free resources held by free ITT queue
 */
static void bnx2i_release_free_itt_queue(struct bnx2i_sess *sess)
{
	sess->itt_q.itt_q_count = 0;
	kfree (sess->itt_q.itt_que_base);
	sess->itt_q.itt_que_base = NULL;

	kfree (sess->itt_q.itt_cmd);
	sess->itt_q.itt_cmd = NULL;
}


/**
 * bnx2i_alloc_cmd - allocates a command structure from free pool
 *
 * @sess:		iscsi session pointer
 *
 * allocates a command structures and ITT from free pool
 */
struct bnx2i_cmd *bnx2i_alloc_cmd(struct bnx2i_sess *sess)
{
	struct bnx2i_cmd *cmd;
	struct list_head *listp;

	if (unlikely(!sess || (sess->num_free_cmds == 0))) {
		return NULL;
	}

	listp = (struct list_head *) sess->free_cmds.next;
	list_del_init(listp);
	sess->num_free_cmds--;
	cmd = (struct bnx2i_cmd *) listp;
	cmd->scsi_status_rcvd = 0;

	bnx2i_setup_cmd_wqe_template(cmd);

	cmd->req.itt = bnx2i_alloc_itt(sess, cmd);

	return cmd;
}


/**
 * bnx2i_free_cmd - releases iscsi cmd struct & ITT to respective free pool
 *
 * @sess:		iscsi session pointer
 * @cmd:		iscsi cmd pointer
 *
 * return command structure and ITT back to free pool.
 */
void bnx2i_free_cmd(struct bnx2i_sess *sess, struct bnx2i_cmd *cmd)
{
	bnx2i_free_itt(sess, cmd);
	list_del_init(&cmd->link);
	list_add_tail(&cmd->link, &sess->free_cmds);
	sess->num_free_cmds++;
}


/**
 * bnx2i_alloc_cmd_pool - allocates and initializes iscsi command pool
 *
 * @sess:		iscsi session pointer
 *
 * Allocate command structure pool for a given iSCSI session. Return 'ENOMEM'
 *	if memory allocation fails
 */
int bnx2i_alloc_cmd_pool(struct bnx2i_sess *sess)
{
	struct bnx2i_cmd *cmdp;
	int index, count;
	int ret_val = 0;
	int total_cmds;
	int num_cmds;
	int page_count = 0;
	int num_cmds_per_page;
	void *mem_ptr;

	INIT_LIST_HEAD(&sess->free_cmds);
	for (index = 0; index < MAX_PAGES_PER_CTRL_STRUCT_POOL; index++)
		sess->cmd_pages[index] = NULL;

	num_cmds_per_page = PAGE_SIZE / sizeof(struct bnx2i_cmd);
	total_cmds = sess->hba->scsi_template->can_queue +
		     BNX2I_RESERVED_SLOW_PATH_CMD_SLOTS;
	if (total_cmds >
	    (num_cmds_per_page * MAX_PAGES_PER_CTRL_STRUCT_POOL))
		total_cmds = num_cmds_per_page *
			     MAX_PAGES_PER_CTRL_STRUCT_POOL;

	for (index = 0; index < total_cmds;) {
		mem_ptr = kmalloc(PAGE_SIZE, GFP_KERNEL);
		if (mem_ptr == NULL)
			break;

		sess->cmd_pages[page_count++] = mem_ptr;
		num_cmds = num_cmds_per_page;
		if ((total_cmds - index) < num_cmds_per_page)
			num_cmds = (total_cmds - index);

		memset(mem_ptr, 0, PAGE_SIZE);
		cmdp = mem_ptr;
		for (count = 0; count < num_cmds; count++) {
			cmdp->req.itt = ITT_INVALID_SIGNATURE;
			INIT_LIST_HEAD(&cmdp->link);

			/* Allocate BD table */
			cmdp->bd_tbl = bnx2i_alloc_bd_table(sess, cmdp);
			if (!cmdp->bd_tbl) {
				/* should never fail, as it's guaranteed to have
				 * (ISCSI_MAX_CMDS_PER_SESS + 1) BD tables
				 * allocated before calling this function.
				 */
				printk(KERN_ERR "no BD table cmd %p\n", cmdp);
				goto bd_table_failed;
			}
			list_add_tail(&cmdp->link, &sess->free_cmds);
			cmdp++;
		}

		sess->num_free_cmds += num_cmds;
		index += num_cmds;
	}
	sess->allocated_cmds = sess->num_free_cmds;

	if (sess->num_free_cmds == 0)
		ret_val = -ENOMEM;
	return ret_val;

bd_table_failed:
	return -ENOMEM;
}


/**
 * bnx2i_free_cmd_pool - releases memory held by free iscsi cmd pool
 *
 * @sess:		iscsi session pointer
 *
 * Release memory held by command struct pool.
 */
void bnx2i_free_cmd_pool(struct bnx2i_sess *sess)
{
	int index;
	void *mem_ptr;

	if (sess->num_free_cmds != sess->allocated_cmds) {
		/*
		 * WARN: either there is some command struct leak or
		 * still some SCSI commands are pending.
		 */
		printk(KERN_ERR "bnx2i: missing cmd structs - %d, %d\n",
				 sess->num_free_cmds, sess->allocated_cmds);
	}
	for (index = 0; index < MAX_PAGES_PER_CTRL_STRUCT_POOL; index++) {
		mem_ptr = sess->cmd_pages[index];
		kfree(mem_ptr);
		sess->cmd_pages[index] = NULL;
	}
	sess->num_free_cmds = sess->allocated_cmds = 0;
	return;
}


/**
 * bnx2i_alloc_bd_table - Alloc BD table to associate with this iscsi cmd 
 *
 * @sess:		iscsi session pointer
 * @cmd:		iscsi cmd pointer
 *
 * allocates a BD table and assigns it to given command structure. There is
 *	no synchronization issue as this code is executed in initialization
 *	thread
 */
static struct io_bdt *bnx2i_alloc_bd_table(struct bnx2i_sess *sess,
					   struct bnx2i_cmd *cmd)
{
	struct io_bdt *bd_tbl;

	if (list_empty(&sess->bd_tbl_list))
		return NULL;

	bd_tbl = (struct io_bdt *)sess->bd_tbl_list.next;
	list_del(&bd_tbl->link);
	list_add_tail(&bd_tbl->link, &sess->bd_tbl_active);
	bd_tbl->bd_valid = 0;
	bd_tbl->cmdp = cmd;

	return bd_tbl;
}


/**
 * bnx2i_free_all_bdt_resc_pages - releases memory held by BD memory tracker tbl
 *
 * @sess:		iscsi session pointer
 *
 * Free up memory pages allocated held by BD resources
 */
static void bnx2i_free_all_bdt_resc_pages(struct bnx2i_sess *sess)
{
	int i;
	struct bd_resc_page *resc_page;

	spin_lock_bh(&sess->lock);
	while (!list_empty(&sess->bd_resc_page)) {
		resc_page = (struct bd_resc_page *)sess->bd_resc_page.prev;
		list_del(sess->bd_resc_page.prev);
		for(i = 0; i < resc_page->num_valid; i++)
			kfree(resc_page->page[i]);
		kfree(resc_page);
	}
	spin_unlock_bh(&sess->lock);
}



/**
 * bnx2i_alloc_bdt_resc_page - allocated a page to track BD table memory
 *
 * @sess:		iscsi session pointer
 *
 * allocated a page to track BD table memory
 */
struct bd_resc_page *bnx2i_alloc_bdt_resc_page(struct bnx2i_sess *sess)
{
	void *mem_ptr;
	struct bd_resc_page *resc_page;

	mem_ptr = kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (!mem_ptr)
		return NULL;

	resc_page = mem_ptr;
	list_add_tail(&resc_page->link, &sess->bd_resc_page);
	resc_page->max_ptrs = (PAGE_SIZE -
		(u32) &((struct bd_resc_page *) 0)->page[0]) / sizeof(void *);
	resc_page->num_valid = 0;

	return resc_page;
}


/**
 * bnx2i_add_bdt_resc_page - add newly allocated memory page to list
 *
 * @sess:		iscsi session pointer
 * @bd_page:		pointer to page memory
 *
 * link newly allocated memory page to tracker list
 */
int bnx2i_add_bdt_resc_page(struct bnx2i_sess *sess, void *bd_page)
{
	struct bd_resc_page *resc_page;

#define is_resc_page_full(_resc_pg) (_resc_pg->num_valid == _resc_pg->max_ptrs)
#define active_resc_page(_resc_list) 	\
			(list_empty(_resc_list) ? NULL : (_resc_list)->prev)
	if (list_empty(&sess->bd_resc_page)) {
		resc_page = bnx2i_alloc_bdt_resc_page(sess);
	} else {
		resc_page = (struct bd_resc_page *)
					active_resc_page(&sess->bd_resc_page);
	}

	if (!resc_page)
		return -ENOMEM;

	resc_page->page[resc_page->num_valid++] = bd_page;
	if (is_resc_page_full(resc_page)) {
		resc_page = bnx2i_alloc_bdt_resc_page(sess);
	}
	return 0;
}


/**
 * bnx2i_alloc_bd_table_pool - Allocates buffer descriptor (BD) pool
 *
 * @sess:		iscsi session pointer
 *
 * Allocate a pool of buffer descriptor tables and associated DMA'able memory
 *	to be used with the session.
 */
static int bnx2i_alloc_bd_table_pool(struct bnx2i_sess *sess)
{
	int index, count;
	int ret_val = 0;
	int num_elem_per_page;
	struct io_bdt *bdt_info;
	void *mem_ptr;
	u32 bd_tbl_size;
	u32 mem_size;
	int total_bd_tbl;
/* check */

	INIT_LIST_HEAD(&sess->bd_resc_page);
	INIT_LIST_HEAD(&sess->bd_tbl_list);
	INIT_LIST_HEAD(&sess->bd_tbl_active);
	total_bd_tbl = sess->hba->scsi_template->can_queue +
		     BNX2I_RESERVED_SLOW_PATH_CMD_SLOTS;
	mem_size = total_bd_tbl * sizeof(struct io_bdt);
	num_elem_per_page = PAGE_SIZE / sizeof(struct io_bdt);
	for (index = 0; index < total_bd_tbl; index += num_elem_per_page) {
		if (((total_bd_tbl - index) * sizeof(struct io_bdt))
		    >= PAGE_SIZE) {
			mem_size = PAGE_SIZE;
			num_elem_per_page = PAGE_SIZE / sizeof(struct io_bdt);
		} else {
			mem_size =
				(total_bd_tbl - index) * sizeof(struct io_bdt);
			num_elem_per_page = (total_bd_tbl - index);
		}
		mem_ptr = kmalloc(mem_size, GFP_KERNEL);
		if (mem_ptr == NULL) {
			printk(KERN_ERR "alloc_bd_tbl: mem alloc failed\n");
			ret_val = -ENOMEM;
			goto resc_alloc_failed;
		}
		bnx2i_add_bdt_resc_page(sess, mem_ptr);

		memset(mem_ptr, 0, mem_size);
		bdt_info = mem_ptr;
		for (count = 0; count < num_elem_per_page; count++) {
			INIT_LIST_HEAD(&bdt_info->link);
			list_add_tail(&bdt_info->link, &sess->bd_tbl_list);
			bdt_info++;
		}
	}

	bd_tbl_size = ISCSI_MAX_BDS_PER_CMD * sizeof(struct iscsi_bd);
	bdt_info = (struct io_bdt *)sess->bd_tbl_list.next;
	while (bdt_info && (bdt_info != (struct io_bdt *)&sess->bd_tbl_list)) {
		mem_ptr = pci_alloc_consistent(sess->hba->pcidev,
					       bd_tbl_size,
					       &bdt_info->bd_tbl_dma);
		if (!mem_ptr) {
			printk(KERN_ERR "bd_tbl: DMA mem alloc failed\n");
			ret_val = -ENOMEM;
			goto dma_alloc_failed;
		}
		bdt_info->bd_tbl = mem_ptr;
		bdt_info->max_bd_cnt = ISCSI_MAX_BDS_PER_CMD;
		bdt_info->bd_valid = 0;
		bdt_info->cmdp = NULL;

		bdt_info = (struct io_bdt *)bdt_info->link.next;
	}
	return ret_val;

resc_alloc_failed:
dma_alloc_failed:
	return ret_val;
}


/**
 * bnx2i_free_bd_table_pool - releases resources held by BD table pool
 *
 * @sess:		iscsi session pointer
 *
 * releases BD table pool memory
 */
void bnx2i_free_bd_table_pool(struct bnx2i_sess *sess)
{
	struct list_head *list;
	struct io_bdt *bdt_info;
	u32 bd_tbl_size;

	bd_tbl_size = ISCSI_MAX_BDS_PER_CMD * sizeof(struct iscsi_bd);
	list_for_each(list, &sess->bd_tbl_list) {
		bdt_info = list_entry(list, struct io_bdt, link);
		pci_free_consistent(sess->hba->pcidev, bd_tbl_size,
				    bdt_info->bd_tbl, bdt_info->bd_tbl_dma);
		bdt_info->bd_tbl = NULL;
		if (bdt_info->cmdp) {
			bdt_info->cmdp->bd_tbl = NULL;
			bdt_info->cmdp = NULL;
		}
	}

	list_for_each(list, &sess->bd_tbl_active) {
		bdt_info = list_entry(list, struct io_bdt, link);
		pci_free_consistent(sess->hba->pcidev, bd_tbl_size,
				    bdt_info->bd_tbl, bdt_info->bd_tbl_dma);
		bdt_info->bd_tbl = NULL;
		if (bdt_info->cmdp) {
			bdt_info->cmdp->bd_tbl = NULL;
			bdt_info->cmdp = NULL;
		}
	}
}


/**
 * bnx2i_setup_mp_bdt - allocated BD table resources to be used as
 *			the dummy buffer for '0' payload length iscsi requests
 *
 * @hba: 		pointer to adapter structure
 *
 * allocate memory for dummy buffer and associated BD table to be used by
 *	middle path (MP) requests
 */
int bnx2i_setup_mp_bdt(struct bnx2i_hba *hba)
{
	int rc = 0;
	struct iscsi_bd *mp_bdt;
	u64 addr;

	hba->mp_bd_tbl = pci_alloc_consistent(hba->pcidev,
					      PAGE_SIZE, &hba->mp_bd_dma);
	if (!hba->mp_bd_tbl) {
		printk(KERN_ERR "unable to allocate Middle Path BDT\n");
		rc = -1;
		goto out;
	}

	hba->dummy_buffer =
		pci_alloc_consistent(hba->pcidev,
				     PAGE_SIZE, &hba->dummy_buf_dma);
	if (!hba->dummy_buffer) {
		printk(KERN_ERR "unable to alloc Middle Path Dummy Buffer\n");
		pci_free_consistent(hba->pcidev, PAGE_SIZE,
				    hba->mp_bd_tbl, hba->mp_bd_dma);
		hba->mp_bd_tbl = NULL;
		rc = -1;
		goto out;
	}

	mp_bdt = (struct iscsi_bd *) hba->mp_bd_tbl;
	addr = (unsigned long) hba->dummy_buf_dma;
	mp_bdt->buffer_addr_lo = addr & 0xffffffff;
	mp_bdt->buffer_addr_hi = addr >> 32;
	mp_bdt->buffer_length = PAGE_SIZE;
	mp_bdt->flags = ISCSI_BD_LAST_IN_BD_CHAIN |
			ISCSI_BD_FIRST_IN_BD_CHAIN;
out:
	return rc;
}


/**
 * bnx2i_free_mp_bdt - releases ITT back to free pool
 *
 * @hba: 		pointer to adapter instance
 *
 * free MP dummy buffer and associated BD table
 */
void bnx2i_free_mp_bdt(struct bnx2i_hba *hba)
{
	if (hba->mp_bd_tbl) {
		pci_free_consistent(hba->pcidev, PAGE_SIZE,
				    hba->mp_bd_tbl, hba->mp_bd_dma);
		hba->mp_bd_tbl = NULL;
	}
	if (hba->dummy_buffer) {
		pci_free_consistent(hba->pcidev, PAGE_SIZE,
				    hba->dummy_buffer, hba->dummy_buf_dma);
		hba->dummy_buffer = NULL;
	}
		return;
}


/**
 * bnx2i_start_iscsi_hba_shutdown - start hba shutdown by cleaning up
 *			all active sessions
 *
 * @hba: 		pointer to adapter instance
 *
 *  interface is being brought down by the user, fail all active iSCSI sessions
 *	belonging to this adapter
 */
void bnx2i_start_iscsi_hba_shutdown(struct bnx2i_hba *hba)
{
	struct list_head *list;
	struct list_head *tmp;
	struct bnx2i_sess *sess;
	int timeout_secs = 30 * HZ;
	unsigned int flags;
	int lpcnt;
	int rc;

	spin_lock_irqsave(&hba->lock, flags);
	list_for_each_safe(list, tmp, &hba->active_sess) {
		sess = (struct bnx2i_sess *)list;
		lpcnt = 4;
		rc = bnx2i_do_iscsi_sess_recovery(sess, DID_NO_CONNECT);
		while ((rc != SUCCESS) && lpcnt--) {
			msleep(1000);
			rc = bnx2i_do_iscsi_sess_recovery(sess, DID_NO_CONNECT);
		}
	}
	spin_unlock_irqrestore(&hba->lock, flags);

	if (test_bit(BNX2I_NX2_DEV_57710, &hba->cnic_dev_type))
		timeout_secs = 240 * HZ;

	 wait_event_interruptible_timeout(hba->eh_wait,
					 (hba->ofld_conns_active == 0),
					 timeout_secs);
}


/**
 * bnx2i_iscsi_handle_ip_event - inetdev callback to handle ip address change
 *
 * @hba: 		pointer to adapter instance
 *
 * IP address change indication, fail all iSCSI connections on this adapter
 *	and let 'iscsid' reinstate the connections
 */
void bnx2i_iscsi_handle_ip_event(struct bnx2i_hba *hba)
{
	struct list_head *list;
	struct list_head *tmp;
	struct bnx2i_sess *sess;
	u32 flags;

	spin_lock_irqsave(&hba->lock, flags);
	list_for_each_safe(list, tmp, &hba->active_sess) {
		sess = (struct bnx2i_sess *)list;
		spin_unlock_irqrestore(&hba->lock, flags);
		bnx2i_do_iscsi_sess_recovery(sess, DID_RESET);
		spin_lock_irqsave(&hba->lock, flags);
	}
	spin_unlock_irqrestore(&hba->lock, flags);
}


/**
 * conn_err_recovery_task - does recovery on all queued sessions
 *
 * @work:		pointer to work struct
 *
 * iSCSI Session recovery queue manager
 */
static void
#if defined(INIT_DELAYED_WORK_DEFERRABLE) || defined(INIT_WORK_NAR)
conn_err_recovery_task(struct work_struct *work)
#else
conn_err_recovery_task(void *data)
#endif
{
#if defined(INIT_DELAYED_WORK_DEFERRABLE) || defined(INIT_WORK_NAR)
	struct bnx2i_hba *hba = container_of(work, struct bnx2i_hba,
					     err_rec_task);
#else
	struct bnx2i_hba *hba = data;
#endif
	struct bnx2i_sess *sess;
	int cons_idx = hba->sess_recov_cons_idx;

	while (hba->sess_recov_prod_idx != cons_idx) {
		sess = hba->sess_recov_list[cons_idx];
		if (sess->state == BNX2I_SESS_IN_LOGOUT)
			bnx2i_do_iscsi_sess_recovery(sess, DID_NO_CONNECT);
		else
			bnx2i_do_iscsi_sess_recovery(sess, DID_RESET);
		if (cons_idx == hba->sess_recov_max_idx)
			cons_idx = 0;
		else
			cons_idx++;
	}
	hba->sess_recov_cons_idx = cons_idx;
}

/**
 * bnx2i_init_ctx_dump_mem - allocate iscsi context dump buffer
 *
 * @hba: 		pointer to adapter instance
 *
 * allocate memory buffer to stage iscsi conn context from the chip
 */
void bnx2i_init_ctx_dump_mem(struct bnx2i_hba *hba)
{
	if (hba->ctx_addr)
		return;

	hba->ictx_poll_mode = 0;
	hba->ctx_size = 0;
	hba->ctx_read_cnt = 0xffffffff;
	hba->ctx_addr = pci_alloc_consistent(hba->pcidev,
					     BNX2I_CONN_CTX_BUF_SIZE,
					     &hba->ctx_dma_hndl);
	if (!hba->ctx_addr)
		return;
	hba->ctx_size = BNX2I_CONN_CTX_BUF_SIZE;
}


/**
 * bnx2i_free_ctx_dump_mem - free context dump memory
 *
 * @hba: 		pointer to adapter instance
 *
 * free context memory buffer
 */
void bnx2i_free_ctx_dump_mem(struct bnx2i_hba *hba)
{
	if (!hba->ctx_addr || (hba->ctx_size == 0))
		return;

	pci_free_consistent(hba->pcidev, hba->ctx_size,
			    hba->ctx_addr, hba->ctx_dma_hndl);
	hba->ctx_dma_hndl = 0;
	hba->ctx_addr = NULL;
	hba->ctx_size = 0;
}


/**
 * bnx2i_ep_destroy_list_add - add an entry to EP destroy list
 *
 * @hba: 		pointer to adapter instance
 * @ep: 		pointer to endpoint (transport indentifier) structure
 *
 * EP destroy queue manager
 */
static int bnx2i_ep_destroy_list_add(struct bnx2i_hba *hba,
				  struct bnx2i_endpoint *ep)
{
	write_lock(&hba->ep_rdwr_lock);
	list_add_tail(&ep->link, &hba->ep_destroy_list);
	write_unlock(&hba->ep_rdwr_lock);
	return 0;
}

/**
 * bnx2i_ep_destroy_list_del - add an entry to EP destroy list
 *
 * @hba: 		pointer to adapter instance
 * @ep: 		pointer to endpoint (transport indentifier) structure
 *
 * EP destroy queue manager
 */
static int bnx2i_ep_destroy_list_del(struct bnx2i_hba *hba,
				     struct bnx2i_endpoint *ep)
{
	write_lock(&hba->ep_rdwr_lock);
	list_del_init(&ep->link);
	write_unlock(&hba->ep_rdwr_lock);

	return 0;
}

/**
 * bnx2i_ep_ofld_list_add - add an entry to ep offload pending list
 *
 * @hba: 		pointer to adapter instance
 * @ep: 		pointer to endpoint (transport indentifier) structure
 *
 * pending conn offload completion queue manager
 */
static int bnx2i_ep_ofld_list_add(struct bnx2i_hba *hba,
				  struct bnx2i_endpoint *ep)
{
	write_lock(&hba->ep_rdwr_lock);
	list_add_tail(&ep->link, &hba->ep_ofld_list);
	write_unlock(&hba->ep_rdwr_lock);
	return 0;
}

/**
 * bnx2i_ep_ofld_list_del - add an entry to ep offload pending list
 *
 * @hba: 		pointer to adapter instance
 * @ep: 		pointer to endpoint (transport indentifier) structure
 *
 * pending conn offload completion queue manager
 */
static int bnx2i_ep_ofld_list_del(struct bnx2i_hba *hba,
				  struct bnx2i_endpoint *ep)
{
	write_lock(&hba->ep_rdwr_lock);
	list_del_init(&ep->link);
	write_unlock(&hba->ep_rdwr_lock);

	return 0;
}


/**
 * bnx2i_find_ep_in_ofld_list - find iscsi_cid in pending list of endpoints
 *
 * @hba: 		pointer to adapter instance
 * @iscsi_cid:		iscsi context ID to find
 *
 */
struct bnx2i_endpoint *
bnx2i_find_ep_in_ofld_list(struct bnx2i_hba *hba, u32 iscsi_cid)
{
	struct list_head *list;
	struct list_head *tmp;
	struct bnx2i_endpoint *ep;

	read_lock(&hba->ep_rdwr_lock);
	list_for_each_safe(list, tmp, &hba->ep_ofld_list) {
		ep = (struct bnx2i_endpoint *)list;

		if (ep->ep_iscsi_cid == iscsi_cid)
			break;
		ep = NULL;
	}
	read_unlock(&hba->ep_rdwr_lock);

	if (!ep)
		printk("bnx2i: ofld_list - icid %d not found\n", iscsi_cid);
	return ep;
}


/**
 * bnx2i_find_ep_in_destroy_list - find iscsi_cid in destroy list
 *
 * @hba: 		pointer to adapter instance
 * @iscsi_cid:		iscsi context ID to find
 *
 */
struct bnx2i_endpoint *
bnx2i_find_ep_in_destroy_list(struct bnx2i_hba *hba, u32 iscsi_cid)
{
	struct list_head *list;
	struct list_head *tmp;
	struct bnx2i_endpoint *ep;

	read_lock(&hba->ep_rdwr_lock);
	list_for_each_safe(list, tmp, &hba->ep_destroy_list) {
		ep = (struct bnx2i_endpoint *)list;

		if (ep->ep_iscsi_cid == iscsi_cid)
			break;
		ep = NULL;
	}
	read_unlock(&hba->ep_rdwr_lock);

	if (!ep)
		printk("bnx2i: destroy_list - icid %d not found\n", iscsi_cid);

	return ep;
}


/**
 * bnx2i_alloc_hba - allocate and init adapter instance
 *
 * @cnic:		cnic device pointer
 *
 * allocate & initialize adapter structure and call other
 *	support routines to do per adapter initialization
 */
struct bnx2i_hba *bnx2i_alloc_hba(struct cnic_dev *cnic)
{
	struct bnx2i_hba *hba;

	hba = kmalloc(sizeof(struct bnx2i_hba), GFP_KERNEL);
	if (!hba)
		return NULL;

	memset(hba, 0, sizeof(struct bnx2i_hba));
	/* Get PCI related information and update hba struct members */
	hba->cnic = cnic;
	hba->netdev = cnic->netdev;

	INIT_LIST_HEAD(&hba->active_sess);
	INIT_LIST_HEAD(&hba->ep_ofld_list);
	INIT_LIST_HEAD(&hba->ep_destroy_list);
	rwlock_init(&hba->ep_rdwr_lock);

	hba->mtu_supported = BNX2I_MAX_MTU_SUPPORTED;

	hba->max_active_conns = ISCSI_MAX_CONNS_PER_HBA;

	if (bnx2i_setup_free_cid_que(hba))
		goto cid_que_err;

	/* Get device type required to determine default SQ size */
	if (cnic->pcidev) {
		hba->pci_did = cnic->pcidev->device;
		bnx2i_identify_device(hba);
	}

	/* SQ/RQ/CQ size can be changed via sysfs interface */
	if (test_bit(BNX2I_NX2_DEV_57710, &hba->cnic_dev_type)) {
		if (sq_size && sq_size <= BNX2I_5770X_SQ_WQES_MAX)
			hba->max_sqes = sq_size;
		else
			hba->max_sqes = BNX2I_5770X_SQ_WQES_DEFAULT;

#ifndef CONFIG_X86_64
		if (hba->max_sqes > BNX2I_5770X_SQ_WQES_DEFAULT_X86)
			hba->max_sqes = BNX2I_5770X_SQ_WQES_DEFAULT_X86;
#endif
	} else {	/* 5706/5708/5709 */
		if (sq_size && sq_size <= BNX2I_570X_SQ_WQES_MAX)
			hba->max_sqes = sq_size;
		else
			hba->max_sqes = BNX2I_570X_SQ_WQES_DEFAULT;
	}

	hba->max_rqes = rq_size;
	hba->max_cqes = hba->max_sqes + rq_size;
	hba->num_ccell = hba->max_sqes / 2;

	spin_lock_init(&hba->lock);
	spin_lock_init(&hba->net_dev_lock);

	/* initialize timer and wait queue used for resource cleanup when
	 * interface is brought down */
	init_timer(&hba->hba_timer);
	init_waitqueue_head(&hba->eh_wait);

#if defined(INIT_DELAYED_WORK_DEFERRABLE) || defined(INIT_WORK_NAR)
	INIT_WORK(&hba->err_rec_task, conn_err_recovery_task);
#else
	INIT_WORK(&hba->err_rec_task, conn_err_recovery_task, hba);
#endif
	hba->sess_recov_prod_idx = 0;
	hba->sess_recov_cons_idx = 0;
	hba->sess_recov_max_idx = 0;
	hba->sess_recov_list = kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (!hba->sess_recov_list)
		goto rec_que_err;
	hba->sess_recov_max_idx = PAGE_SIZE / sizeof (struct bnx2i_sess *) - 1;

	return hba;

rec_que_err:
	bnx2i_release_free_cid_que(hba);
cid_que_err:
	bnx2i_free_hba(hba);

	return NULL;
}


/**
 * bnx2i_free_hba- releases hba structure and resources held by the adapter
 *
 * @hba: 		pointer to adapter instance
 *
 * free adapter structure and call various cleanup routines.
 */
void bnx2i_free_hba(struct bnx2i_hba *hba)
{
	bnx2i_release_free_cid_que(hba);
	INIT_LIST_HEAD(&hba->ep_ofld_list);
	INIT_LIST_HEAD(&hba->ep_destroy_list);

	INIT_LIST_HEAD(&hba->active_sess);
	/* Free memory held by hba structure */
	kfree(hba);
}


/**
 * bnx2i_flush_active_cmd_queue - flush active command queue
 *
 * @sess:		iscsi session pointer
 * @err_code: 		SCSI ML error code, DID_BUS_BUSY
 *
 * return all commands in active queue which should already have been
 * 	cleaned up by the cnic device.
 */
static void bnx2i_flush_active_cmd_queue(struct bnx2i_sess *sess, int err_code)
{
	struct list_head *list;
	struct list_head *tmp;
	struct bnx2i_cmd *cmd;
	unsigned long flags;

	if (!sess->num_active_cmds)
		return;

	spin_lock_irqsave(sess->host->host_lock, flags);
	list_for_each_safe(list, tmp, &sess->active_cmds) {
		cmd = (struct bnx2i_cmd *) list;
		cmd->req.itt &= ISCSI_CMD_RESPONSE_INDEX;
		list_del_init(&cmd->link);
		if (cmd->req.op_code == ISCSI_OP_SCSI_CMD) {
			bnx2i_iscsi_unmap_sg_list(sess->hba, cmd);
			cmd->cmd_state = ISCSI_CMD_STATE_COMPLETED;
			bnx2i_return_failed_command(sess, cmd, err_code);
		} else
			sess->num_active_cmds--;
		bnx2i_free_cmd(sess, cmd);
	}
	spin_unlock_irqrestore(sess->host->host_lock, flags);
}


/**
 * bnx2i_session_recovery_start - start recovery process on given session
 *
 * @sess:		iscsi session pointer
 * @err_code: 		SCSI ML error code, DID_BUS_BUSY
 *
 * initiate cleanup of outstanding commands for sess recovery
 */
static int bnx2i_session_recovery_start(struct bnx2i_sess *sess, int err_code)
{
	if (sess->state == BNX2I_SESS_IN_LOGOUT ||
	    sess->state == BNX2I_SESS_INITIAL)
		return 0;

	if (!is_sess_active(sess) &&
	    !sess->state & BNX2I_SESS_INITIAL) {
		if (sess->recovery_state)
			return -EPERM;
		wait_event_interruptible_timeout(sess->er_wait,
						 (sess->state ==
						  BNX2I_SESS_IN_FFP), 20 * HZ);
		if (signal_pending(current))
			flush_signals(current);
		if (!is_sess_active(sess) &&
		    !sess->state & BNX2I_SESS_INITIAL) {
			printk(KERN_ALERT "sess_reco: sess still not active\n");
			sess->lead_conn->state = CONN_STATE_XPORT_FREEZE;
			return -EPERM;
		}
	}

	return 0;
}


/**
 * bnx2i_do_iscsi_sess_recovery - implements session recovery code
 *
 * @sess:		iscsi session pointer
 * @err_code: 		SCSI ML error code, DID_BUS_BUSY, DID_NO_CONNECT,
 *			DID_RESET
 *
 * SCSI host reset handler, which is translates to iSCSI session
 *	recovery. This routine starts internal driver session recovery,
 *	indicates connection error to 'iscsid' which does session reinstatement
 *	This is an synchronous call which waits for completion and returns
 *	the ultimate result of session recovery process to caller
 */
int bnx2i_do_iscsi_sess_recovery(struct bnx2i_sess *sess, int err_code)
{
	struct bnx2i_hba *hba;
	struct bnx2i_conn *conn = sess->lead_conn;

	if (!conn)
		return FAILED;

	/* block scsi host to avoid any further command queuing */
	iscsi_block_session(session_to_cls(sess));

	if (bnx2i_session_recovery_start(sess, err_code)) {
		printk(KERN_INFO "bnx2i: sess rec start returned error\n");
		return FAILED;
	}
	hba = sess->hba;

	sess->recovery_state = ISCSI_SESS_RECOVERY_OPEN_ISCSI;
	iscsi_conn_error(conn->cls_conn, ISCSI_ERR_CONN_FAILED);

	/* if session teardown is because of net interface down,
	 * no need to wait for complete recovery */
	if (err_code == DID_NO_CONNECT ||
	    test_bit(ADAPTER_STATE_LINK_DOWN, &sess->hba->adapter_state))
		wait_event_interruptible_timeout(sess->er_wait,
						 !conn->ep,
						 msecs_to_jiffies(3000));
	else
		wait_event_interruptible(sess->er_wait,
					 ((sess->recovery_state &
					   ISCSI_SESS_RECOVERY_COMPLETE) ||
					  (sess->recovery_state &
					   ISCSI_SESS_RECOVERY_FAILED)));

	if (signal_pending(current))
		flush_signals(current);

	if (err_code == DID_NO_CONNECT)
		goto ret_success;

	if (sess->recovery_state & ISCSI_SESS_RECOVERY_COMPLETE) {
		printk(KERN_INFO "bnx2i: host #%d reset succeeded\n",
				  sess->host->host_no);
		sess->state = BNX2I_SESS_IN_FFP;
	} else
		return FAILED;

ret_success:
	sess->recovery_state = 0;
	return SUCCESS;
}


/**
 * bnx2i_iscsi_sess_release - cleanup iscsi session & reclaim all resources
 *
 * @hba: 		pointer to adapter instance
 * @sess:		iscsi session pointer
 *
 * free up resources held by this session including ITT queue, cmd struct pool,
 *	BD table pool. HBA lock is held while manipulating active session list
 */
void bnx2i_iscsi_sess_release(struct bnx2i_hba *hba, struct bnx2i_sess *sess)
{
	u32 flags;

	bnx2i_release_free_itt_queue(sess);
	bnx2i_free_bd_table_pool(sess);
	bnx2i_free_all_bdt_resc_pages(sess);
	bnx2i_free_cmd_pool(sess);

	spin_lock_irqsave(&hba->lock, flags);
	list_del_init(&sess->link);
	hba->num_active_sess--;
	spin_unlock_irqrestore(&hba->lock, flags);
}


/**
 * bnx2i_iscsi_sess_new - initialize newly allocated session structure
 *
 * @hba: 		pointer to adapter instance
 * @sess:		iscsi session pointer
 *
 * initialize session structure elements and allocate per sess resources.
 *	Some of the per session resources allocated are command struct pool,
 *	BD table pool and ITT queue region
 */
int bnx2i_iscsi_sess_new(struct bnx2i_hba *hba, struct bnx2i_sess *sess)
{
	u32 flags;

	spin_lock_irqsave(&hba->lock, flags);
	list_add_tail(&sess->link, &hba->active_sess);
	hba->num_active_sess++;
	spin_unlock_irqrestore(&hba->lock, flags);

	sess->sq_size = hba->max_sqes;
	sess->tsih = 0;
	sess->lead_conn = NULL;

	spin_lock_init(&sess->lock);

	/* initialize active connection list */
	INIT_LIST_HEAD(&sess->conn_list);
	INIT_LIST_HEAD(&sess->free_cmds);

	INIT_LIST_HEAD(&sess->active_cmds);
	sess->num_active_cmds = 0;

	sess->num_active_conn = 0;
	sess->max_conns = 1;
	sess->target_name = NULL;

	sess->state = BNX2I_SESS_INITIAL;
	sess->recovery_state = 0;
	sess->old_recovery_state = 0;
	sess->tmf_active = 0;

	if (bnx2i_alloc_bd_table_pool(sess) != 0) {
		printk(KERN_ERR "sess_new: unable to alloc bd table pool\n");
		goto err_bd_pool;
	}

	if (bnx2i_alloc_cmd_pool(sess) != 0) {
		printk(KERN_ERR "sess_new: alloc cmd pool failed\n");
		goto err_cmd_pool;
	}

	if (bnx2i_setup_free_itt_queue(sess)) {
		printk(KERN_ERR "sess_new: unable to alloc itt queue\n");
		goto err_itt_que;
	}

	init_timer(&sess->abort_timer);
	init_waitqueue_head(&sess->er_wait);

	return 0;

err_itt_que:
	bnx2i_free_cmd_pool(sess);
err_cmd_pool:
	bnx2i_free_bd_table_pool(sess);
err_bd_pool:
	return -ENOMEM;
}

/**
 * bnx2i_conn_free_login_resources - free DMA resources used for login process
 *
 * @hba: 		pointer to adapter instance
 * @conn: 		iscsi connection pointer
 *
 * Login related resources, mostly BDT & payload DMA memory is freed
 */
void bnx2i_conn_free_login_resources(struct bnx2i_hba *hba,
				     struct bnx2i_conn *conn)
{
	if (conn->gen_pdu.resp_bd_tbl) {
		pci_free_consistent(hba->pcidev, PAGE_SIZE,
				    conn->gen_pdu.resp_bd_tbl,
				    conn->gen_pdu.resp_bd_dma);
		conn->gen_pdu.resp_bd_tbl = NULL;
	}

	if (conn->gen_pdu.req_bd_tbl) {
		pci_free_consistent(hba->pcidev, PAGE_SIZE,
				    conn->gen_pdu.req_bd_tbl,
				    conn->gen_pdu.req_bd_dma);
		conn->gen_pdu.req_bd_tbl = NULL;
	}

	if (conn->gen_pdu.resp_buf) {
		pci_free_consistent(hba->pcidev, ISCSI_CONN_LOGIN_BUF_SIZE,
				    conn->gen_pdu.resp_buf,
				    conn->gen_pdu.resp_dma_addr);
		conn->gen_pdu.resp_buf = NULL;
	}

	if (conn->gen_pdu.req_buf) {
		pci_free_consistent(hba->pcidev, ISCSI_CONN_LOGIN_BUF_SIZE,
				    conn->gen_pdu.req_buf,
				    conn->gen_pdu.req_dma_addr);
		conn->gen_pdu.req_buf = NULL;
	}
}

/**
 * bnx2i_conn_alloc_login_resources - alloc DMA resources used for
 *			login / nopout pdus
 *
 * @hba: 		pointer to adapter instance
 * @conn: 		iscsi connection pointer
 *
 * Login & nop-in related resources is allocated in this routine.
 */
static int bnx2i_conn_alloc_login_resources(struct bnx2i_hba *hba,
					    struct bnx2i_conn *conn)
{
	/* Allocate memory for login request/response buffers */
	conn->gen_pdu.req_buf =
		pci_alloc_consistent(hba->pcidev, ISCSI_CONN_LOGIN_BUF_SIZE,
				     &conn->gen_pdu.req_dma_addr);
	if (conn->gen_pdu.req_buf == NULL)
		goto login_req_buf_failure;

	conn->gen_pdu.req_buf_size = 0;
	conn->gen_pdu.req_wr_ptr = conn->gen_pdu.req_buf;

	conn->gen_pdu.resp_buf =
		pci_alloc_consistent(hba->pcidev, ISCSI_CONN_LOGIN_BUF_SIZE,
				     &conn->gen_pdu.resp_dma_addr);
	if (conn->gen_pdu.resp_buf == NULL)
		goto login_resp_buf_failure;

	conn->gen_pdu.resp_buf_size = ISCSI_CONN_LOGIN_BUF_SIZE;
	conn->gen_pdu.resp_wr_ptr = conn->gen_pdu.resp_buf;
	
	conn->gen_pdu.req_bd_tbl =
		pci_alloc_consistent(hba->pcidev, PAGE_SIZE,
				     &conn->gen_pdu.req_bd_dma);
	if (conn->gen_pdu.req_bd_tbl == NULL)
		goto login_req_bd_tbl_failure;

	conn->gen_pdu.resp_bd_tbl =
		pci_alloc_consistent(hba->pcidev, PAGE_SIZE,
				     &conn->gen_pdu.resp_bd_dma);
	if (conn->gen_pdu.resp_bd_tbl == NULL)
		goto login_resp_bd_tbl_failure;

	return 0;

login_resp_bd_tbl_failure:
	pci_free_consistent(hba->pcidev, PAGE_SIZE, conn->gen_pdu.req_bd_tbl,
			    conn->gen_pdu.req_bd_dma);
	conn->gen_pdu.req_bd_tbl = NULL;

login_req_bd_tbl_failure:
	pci_free_consistent(hba->pcidev, ISCSI_CONN_LOGIN_BUF_SIZE,
			    conn->gen_pdu.resp_buf,
			    conn->gen_pdu.resp_dma_addr);
	conn->gen_pdu.resp_buf = NULL;
login_resp_buf_failure:
	pci_free_consistent(hba->pcidev, ISCSI_CONN_LOGIN_BUF_SIZE,
			    conn->gen_pdu.req_buf, conn->gen_pdu.req_dma_addr);
	conn->gen_pdu.req_buf = NULL;
login_req_buf_failure:
	printk(KERN_ERR "bnx2i:a conn login resource alloc failed!!\n");
	return -ENOMEM;

}


/**
 * bnx2i_iscsi_conn_new - initialize newly created connection structure
 *
 * @sess:		iscsi session pointer
 * @conn: 		iscsi connection pointer
 *
 * connection structure is initialized which mainly includes allocation of
 *	login resources and lock/time initialization
 */
int bnx2i_iscsi_conn_new(struct bnx2i_sess *sess, struct bnx2i_conn *conn)
{
	struct bnx2i_hba *hba = sess->hba;

	conn->sess = sess;
	conn->header_digest_en = 0;
	conn->data_digest_en = 0;

	spin_lock_init(&conn->lock);
	INIT_LIST_HEAD(&conn->link);

	conn->gen_pdu.cmd = NULL;

	/* 'ep' ptr will be assigned in bind() call */
	conn->ep = NULL;

	if (bnx2i_conn_alloc_login_resources(hba, conn)) {
		printk(KERN_ALERT "conn_new: login resc alloc failed!!\n");
		return -ENOMEM;
	}

	return 0;
}


/**
 * bnx2i_login_resp_update_cmdsn - extracts SN & MAX_SN from login response header &
 *			updates driver 'cmdsn' with 
 *
 * @conn: 		iscsi connection pointer
 *
 * extract & update SN counters from login response
 */
static int bnx2i_login_resp_update_cmdsn(struct bnx2i_conn *conn)
{
	u32 max_cmdsn;
	u32 exp_cmdsn;
	u32 stat_sn;
	struct bnx2i_sess *sess = conn->sess;
	struct iscsi_nopin *hdr;

	hdr = (struct iscsi_nopin *) &conn->gen_pdu.resp_hdr;

	max_cmdsn = ntohl(hdr->max_cmdsn);
	exp_cmdsn = ntohl(hdr->exp_cmdsn);
	stat_sn = ntohl(hdr->statsn);
#define SN_DELTA_ISLAND		0xffff
	if (max_cmdsn < exp_cmdsn -1 &&
	    max_cmdsn > exp_cmdsn - SN_DELTA_ISLAND)
		return -EINVAL;

	if (max_cmdsn > sess->max_cmdsn ||
	    max_cmdsn < sess->max_cmdsn - SN_DELTA_ISLAND)
		sess->max_cmdsn = max_cmdsn;

	if (exp_cmdsn > sess->exp_cmdsn ||
	    exp_cmdsn < sess->exp_cmdsn - SN_DELTA_ISLAND)
		sess->exp_cmdsn = exp_cmdsn;

	if (stat_sn == conn->exp_statsn)
		conn->exp_statsn++;

	return 0;
}


/**
 * bnx2i_update_cmd_sequence - update session sequencing parameter
 *
 * @sess:		iscsi session pointer
 * @exp_sn: 		iscsi expected command seq num
 * @max_sn: 		iscsi max command seq num
 *
 * update iSCSI SN counters for the given session
 */
void bnx2i_update_cmd_sequence(struct bnx2i_sess *sess,
			       u32 exp_sn, u32 max_sn)
{
	u32 exp_cmdsn = exp_sn;
	u32 max_cmdsn = max_sn;

	if (max_cmdsn < exp_cmdsn -1 &&
	    max_cmdsn > exp_cmdsn - SN_DELTA_ISLAND) {
		printk(KERN_ALERT "cmd_sequence: error, exp 0x%x, max 0x%x\n",
				   exp_cmdsn, max_cmdsn);
		BUG_ON(1);
	}
	if (max_cmdsn > sess->max_cmdsn ||
	    max_cmdsn < sess->max_cmdsn - SN_DELTA_ISLAND)
		sess->max_cmdsn = max_cmdsn;
	if (exp_cmdsn > sess->exp_cmdsn ||
	    exp_cmdsn < sess->exp_cmdsn - SN_DELTA_ISLAND)
		sess->exp_cmdsn = exp_cmdsn;
}


/**
 * bnx2i_process_scsi_resp - complete SCSI command processing by calling
 *			'scsi_done', free iscsi cmd structure to free list
 *
 * @cmd:		iscsi cmd pointer
 * @resp_cqe:		scsi response cqe pointer
 *
 * validates scsi response indication for normal completion, sense data if any
 *	underflow/overflow condition and propogates SCSI response to SCSI-ML by
 *	calling scsi_done() and also returns command struct back to free pool
 */
void bnx2i_process_scsi_resp(struct bnx2i_cmd *cmd,
			    struct iscsi_cmd_response *resp_cqe)
{
	struct scsi_cmnd *sc = cmd->scsi_cmd;
	struct Scsi_Host *host;
	u16 sense_data[128];
	int data_len;
	u16 sense_len;

	host = cmd->conn->sess->host;
	sc->result = (DID_OK << 16) | resp_cqe->status;

	if (resp_cqe->response != ISCSI_STATUS_CMD_COMPLETED) {
		sc->result = (DID_ERROR << 16);
		goto out;
	}

	if (resp_cqe->status == SAM_STAT_CHECK_CONDITION) {
		data_len = resp_cqe->data_length;
		if (data_len < 2) {
			printk(KERN_ERR "bnx2i: CHK_CONDITION - invalid "
					 "data length %d\n", data_len);
			goto out;
		}

		if (data_len > BNX2I_RQ_WQE_SIZE) {
			printk(KERN_ALERT "bnx2i: sense data len %d > RQ sz\n",
					  data_len);
			data_len = BNX2I_RQ_WQE_SIZE;
		}
		if (data_len) {
			memset(sc->sense_buffer, 0, sizeof(sc->sense_buffer));
			bnx2i_get_rq_buf(cmd->conn, (char *)sense_data, data_len);
			bnx2i_put_rq_buf(cmd->conn, 1);
			cmd->conn->total_data_octets_rcvd += data_len;
			sense_len = be16_to_cpu(*((__be16 *) sense_data));

			if (sense_len > SCSI_SENSE_BUFFERSIZE)
				sense_len = SCSI_SENSE_BUFFERSIZE;

			memcpy(sc->sense_buffer, &sense_data[1],
			       (int) sense_len);
		}
	}


	if (sc->sc_data_direction == DMA_TO_DEVICE)
		goto out;
	

	if (resp_cqe->response_flags &
	    ISCSI_CMD_RESPONSE_RESIDUAL_UNDERFLOW) {
		if (resp_cqe->residual_count &&
		    (resp_cqe->residual_count <= scsi_bufflen(sc))) {
			scsi_set_resid(sc, resp_cqe->residual_count);
			cmd->conn->total_data_octets_rcvd -= scsi_get_resid(sc);
		} else
			sc->result = (DID_BAD_TARGET << 16) | resp_cqe->status;
	} else if (resp_cqe->response_flags &
		   ISCSI_CMD_RESPONSE_RESIDUAL_OVERFLOW) {
		scsi_set_resid(sc, resp_cqe->residual_count);
		cmd->conn->total_data_octets_rcvd += scsi_get_resid(sc);
	}

out:
	return;

}

/**
 * bnx2i_indicate_login_resp - process iscsi login response
 *
 * @conn: 		iscsi connection pointer
 *
 * pushes login response PDU to application daemon, 'iscsid' by
 *		calling iscsi_recv_pdu()
 */
int bnx2i_indicate_login_resp(struct bnx2i_conn *conn)
{
	int data_len;
	struct iscsi_login_rsp *login_resp =
		(struct iscsi_login_rsp *) &conn->gen_pdu.resp_hdr;

	/* check if this is the first login response for this connection.
	 * If yes, we need to copy initial StatSN to connection structure.
	 */
	if (conn->exp_statsn == STATSN_UPDATE_SIGNATURE) {
		conn->exp_statsn = ntohl(login_resp->statsn) + 1;
	}

	if (bnx2i_login_resp_update_cmdsn(conn))
		return -EINVAL;

	data_len = conn->gen_pdu.resp_wr_ptr - conn->gen_pdu.resp_buf;
	iscsi_recv_pdu(conn->cls_conn, (struct iscsi_hdr *) login_resp,
		       (char *) conn->gen_pdu.resp_buf, data_len);

	return 0;
}


/**
 * bnx2i_indicate_logout_resp - process iscsi logout response
 *
 * @conn: 		iscsi connection pointer
 *
 * pushes logout response PDU to application daemon, 'iscsid' by
 *		calling iscsi_recv_pdu()
 */
int bnx2i_indicate_logout_resp(struct bnx2i_conn *conn)
{
	unsigned long flags;
	struct iscsi_logout_rsp *logout_resp =
		(struct iscsi_logout_rsp *) &conn->gen_pdu.resp_hdr;

	iscsi_recv_pdu(conn->cls_conn, (struct iscsi_hdr *) logout_resp,
		       (char *) NULL, 0);
	
	if (conn->gen_pdu.cmd) {
		spin_lock_irqsave(conn->sess->host->host_lock, flags);
		list_del_init(&conn->gen_pdu.cmd->link);
		bnx2i_free_cmd(conn->sess, conn->gen_pdu.cmd);
		conn->gen_pdu.cmd = NULL;
		spin_unlock_irqrestore(conn->sess->host->host_lock, flags);
	}
	return 0;
}


/**
 * bnx2i_indicate_async_mesg - process iscsi ASYNC message indication
 *
 * @conn: 		iscsi connection pointer
 *
 * pushes iSCSI async PDU to application daemon, 'iscsid' by calling
 *	iscsi_recv_pdu()
 */
int bnx2i_indicate_async_mesg(struct bnx2i_conn *conn)
{
	struct iscsi_async *async_msg =
		(struct iscsi_async *) &conn->gen_pdu.resp_hdr;

	iscsi_recv_pdu(conn->cls_conn, (struct iscsi_hdr *) async_msg,
		       (char *) NULL, 0);
	return 0;
}



/**
 * bnx2i_process_nopin - process iscsi nopin pdu
 *
 * @conn: 		iscsi connection pointer
 * @cmd:		iscsi cmd pointer
 * @data_buf:		payload buffer pointer
 * @data_len:		payload length
 *
 * pushes nopin pdu to application daemon, 'iscsid' by calling iscsi_recv_pdu
 */
int bnx2i_process_nopin(struct bnx2i_conn *conn, struct bnx2i_cmd *cmd,
			char *data_buf, int data_len)
{
	struct iscsi_nopin *nopin_msg =
		(struct iscsi_nopin *) &conn->gen_pdu.resp_hdr;

	iscsi_recv_pdu(conn->cls_conn, (struct iscsi_hdr *) nopin_msg,
		       (char *) data_buf, data_len);

	spin_lock(conn->sess->host->host_lock);
	list_del_init(&cmd->link);
	conn->sess->num_active_cmds--;
	bnx2i_free_cmd(cmd->conn->sess, cmd);
	spin_unlock(conn->sess->host->host_lock);

	return 0;
}



/**
 * bnx2i_iscsi_prep_generic_pdu_bd - prepares BD table to be used with
 *			generic iscsi pdus
 *
 * @conn: 		iscsi connection pointer
 *
 * Allocates buffers and BD tables before shipping requests to cnic
 *	for PDUs prepared by 'iscsid' daemon
 */
static void bnx2i_iscsi_prep_generic_pdu_bd(struct bnx2i_conn *conn)
{
	struct iscsi_bd *bd_tbl;

	bd_tbl = (struct iscsi_bd *) conn->gen_pdu.req_bd_tbl;

	bd_tbl->buffer_addr_hi =
		(u32) ((u64) conn->gen_pdu.req_dma_addr >> 32);
	bd_tbl->buffer_addr_lo = (u32) conn->gen_pdu.req_dma_addr;
	bd_tbl->buffer_length = conn->gen_pdu.req_wr_ptr -
				conn->gen_pdu.req_buf;
	bd_tbl->reserved0 = 0;
	bd_tbl->flags = ISCSI_BD_LAST_IN_BD_CHAIN |
			ISCSI_BD_FIRST_IN_BD_CHAIN;

	bd_tbl = (struct iscsi_bd  *) conn->gen_pdu.resp_bd_tbl;
	bd_tbl->buffer_addr_hi = (u64) conn->gen_pdu.resp_dma_addr >> 32;
	bd_tbl->buffer_addr_lo = (u32) conn->gen_pdu.resp_dma_addr;
	bd_tbl->buffer_length = ISCSI_CONN_LOGIN_BUF_SIZE;
	bd_tbl->reserved0 = 0;
	bd_tbl->flags = ISCSI_BD_LAST_IN_BD_CHAIN |
			ISCSI_BD_FIRST_IN_BD_CHAIN;
}


/**
 * bnx2i_nopout_check_active_cmds - checks if iscsi link is idle
 *
 * @hba: 		pointer to adapter instance
 *
 * called to check if iscsi connection is idle or not. Pro-active nopout
 *	 is sent only if the link is idle
 */
static int bnx2i_nopout_check_active_cmds(struct bnx2i_conn *conn,
					  struct bnx2i_cmd *cmnd)
{
	struct iscsi_nopin *nopin_msg =
		(struct iscsi_nopin *) &conn->gen_pdu.resp_hdr;
	int len;
	char *buf = NULL;
	unsigned long flags;

	if ((conn->nopout_num_scsi_cmds == conn->num_scsi_cmd_pdus) &&
	    !conn->sess->num_active_cmds) {
		return -1;
	}

	len = conn->gen_pdu.req_buf_size;
	if (len)
		buf = conn->gen_pdu.req_buf;

	memset(nopin_msg, 0x00, sizeof(struct iscsi_nopin));
        nopin_msg->opcode = ISCSI_OP_NOOP_IN;
        nopin_msg->flags = ISCSI_FLAG_CMD_FINAL;
        memcpy(nopin_msg->lun, conn->gen_pdu.pdu_hdr.lun, 8);
        nopin_msg->itt = conn->gen_pdu.pdu_hdr.itt;
        nopin_msg->ttt = ISCSI_RESERVED_TAG;
        nopin_msg->statsn = conn->gen_pdu.pdu_hdr.exp_statsn;;
        nopin_msg->exp_cmdsn = htonl(conn->sess->exp_cmdsn);
        nopin_msg->max_cmdsn = htonl(conn->sess->max_cmdsn);

	iscsi_recv_pdu(conn->cls_conn, (struct iscsi_hdr *) nopin_msg,
		       (char *) buf, len);

	conn->nopout_num_scsi_cmds = conn->num_scsi_cmd_pdus;

	spin_lock_irqsave(conn->sess->host->host_lock, flags);
	list_del_init(&cmnd->link);
	bnx2i_free_cmd(conn->sess, cmnd);
	spin_unlock_irqrestore(conn->sess->host->host_lock, flags);
	return 0;
}


/**
 * bnx2i_iscsi_send_generic_request - called to send iscsi login/nopout/logout
 *			pdus
 *
 * @hba: 		pointer to adapter instance
 *
 * called to transmit PDUs prepared by the 'iscsid' daemon. iSCSI login,
 *	Nop-out and Logout requests flow through this path.
 */
static int bnx2i_iscsi_send_generic_request(struct bnx2i_cmd *cmnd)
{
	int rc = 0;
	char *buf;
	int data_len;
	struct bnx2i_conn *conn = cmnd->conn;

	bnx2i_iscsi_prep_generic_pdu_bd(conn);
	switch (cmnd->iscsi_opcode & ISCSI_OPCODE_MASK) {
	case ISCSI_OP_LOGIN:
		bnx2i_send_iscsi_login(conn, cmnd);
		break;

	case ISCSI_OP_NOOP_OUT:
		if (!bnx2i_nopout_when_cmds_active)
			if (!bnx2i_nopout_check_active_cmds(conn, cmnd)) {
				return 0;
			}

		conn->nopout_num_scsi_cmds = conn->num_scsi_cmd_pdus;
		data_len = conn->gen_pdu.req_buf_size;
		buf = conn->gen_pdu.req_buf;
		if (data_len)
			rc = bnx2i_send_iscsi_nopout(conn, cmnd,
						     ISCSI_RESERVED_TAG,
						     buf, data_len, 1);
		else
			rc = bnx2i_send_iscsi_nopout(conn, cmnd,
						     ISCSI_RESERVED_TAG,
						     NULL, 0, 1);
		break;

	case ISCSI_OP_LOGOUT:
		rc = bnx2i_send_iscsi_logout(conn, cmnd);
		break;

	default:
		printk(KERN_ALERT "send_gen: unsupported op 0x%x\n",
				   cmnd->iscsi_opcode);
	}
	return rc;
}


/**********************************************************************
 *		SCSI-ML Interface
 **********************************************************************/

/**
 * bnx2i_cpy_scsi_cdb - copies LUN & CDB fields in required format to sq wqe
 *
 * @sc: 		SCSI-ML command pointer
 * @cmd:		iscsi cmd pointer
 *
 */
static void bnx2i_cpy_scsi_cdb(struct scsi_cmnd *sc,
				      struct bnx2i_cmd *cmd)
{
	u32 dword;
	int lpcnt;
	u8 *srcp;
	u32 *dstp;
	u32 scsi_lun[2];

	int_to_scsilun(sc->device->lun, (struct scsi_lun *) scsi_lun);
	cmd->req.lun[0] = ntohl(scsi_lun[0]);
	cmd->req.lun[1] = ntohl(scsi_lun[1]);

	lpcnt = cmd->scsi_cmd->cmd_len / sizeof(dword);
	srcp = (u8 *) sc->cmnd;
	dstp = (u32 *) cmd->req.cdb;
	while (lpcnt--) {
		memcpy(&dword, (const void *) srcp, 4);
		*dstp = cpu_to_be32(dword);
		srcp += 4;
		dstp++;
	}
	if (sc->cmd_len & 0x3) {
		dword = (u32) srcp[0] | ((u32) srcp[1] << 8);
		*dstp = cpu_to_be32(dword);
	}
}



/**
 * bnx2i_queuecommand - SCSI ML - bnx2i interface function to issue new commands
*			to be shipped to iscsi target
 *
 * @sc: 		SCSI-ML command pointer
 * @done: 		callback function pointer to complete the task
 *
 * handles SCSI command queued by SCSI-ML, allocates a command structure,
 *	assigning CMDSN, mapping SG buffers and delivers it to CNIC for further
 *	processing. This routine also takes care of iSCSI command window full
 *	condition, if session is in recovery process and other error conditions
 */
int bnx2i_queuecommand(struct scsi_cmnd *sc,
		       void (*done) (struct scsi_cmnd *))
{
	struct bnx2i_sess *sess;
	struct bnx2i_conn *conn;
	struct bnx2i_cmd *cmd;

	sc->scsi_done = done;
	sc->result = 0;
	sess = iscsi_hostdata(sc->device->host->hostdata);

#define iscsi_cmd_win_closed(_sess)	\
	((int) (_sess->max_cmdsn - _sess->cmdsn) < 0)

	if (iscsi_cmd_win_closed(sess))
		goto iscsi_win_closed;

	if (sess->state & BNX2I_SESS_IN_SHUTDOWN ||
	    sess->state & BNX2I_SESS_IN_LOGOUT || !sess->lead_conn)
		goto dev_not_found;

	if (sess->recovery_state) {
		if (sess->old_recovery_state != sess->recovery_state)
			sess->old_recovery_state = sess->recovery_state;

		if (sess->recovery_state & ISCSI_SESS_RECOVERY_FAILED)
			goto dev_not_found;
		else if (!(sess->recovery_state & ISCSI_SESS_RECOVERY_COMPLETE))
			goto iscsi_win_closed;
		else {	/* ISCSI_SESS_RECOVERY_COMPLETE */
			sess->old_recovery_state = sess->recovery_state = 0;
		}
	}

	if (test_bit(ADAPTER_STATE_LINK_DOWN, &sess->hba->adapter_state))
		goto iscsi_win_closed;

	cmd = bnx2i_alloc_cmd(sess);
	if (cmd == NULL)
		/* should never happen as cmd list size == SHT->can_queue */
		goto cmd_not_accepted;

	cmd->conn = conn = sess->lead_conn;
	cmd->scsi_cmd = sc;
	cmd->req.total_data_transfer_length = scsi_bufflen(sc);
	cmd->iscsi_opcode = cmd->req.op_code = ISCSI_OP_SCSI_CMD;
	cmd->req.cmd_sn = sess->cmdsn++;

	bnx2i_iscsi_map_sg_list(cmd);
	bnx2i_cpy_scsi_cdb(sc, cmd);

	cmd->req.op_attr = ISCSI_ATTR_SIMPLE;
	if (sc->sc_data_direction == DMA_TO_DEVICE) {
		cmd->req.op_attr |= ISCSI_CMD_REQUEST_WRITE;
		cmd->req.itt |= (ISCSI_TASK_TYPE_WRITE <<
				 ISCSI_CMD_REQUEST_TYPE_SHIFT);
		bnx2i_setup_write_cmd_bd_info(cmd);
	} else {
		if (scsi_bufflen(sc))
			cmd->req.op_attr |= ISCSI_CMD_REQUEST_READ;
		cmd->req.itt |= (ISCSI_TASK_TYPE_READ <<
				 ISCSI_CMD_REQUEST_TYPE_SHIFT);
	}
	cmd->req.num_bds = cmd->bd_tbl->bd_valid;
	if (!cmd->bd_tbl->bd_valid) {
		cmd->req.bd_list_addr_lo = (u32) sess->hba->mp_bd_dma;
		cmd->req.bd_list_addr_hi =
			(u32) ((u64) sess->hba->mp_bd_dma >> 32);
		cmd->req.num_bds = 1;
	}

	cmd->cmd_state = ISCSI_CMD_STATE_INITIATED;
	sc->SCp.ptr = (char *) cmd;

	if (cmd->req.itt != ITT_INVALID_SIGNATURE) {
		bnx2i_send_iscsi_scsicmd(conn, cmd);
		list_add_tail(&cmd->link, &sess->active_cmds);
		sess->num_active_cmds++;
	}
	return 0;

iscsi_win_closed:
cmd_not_accepted:
	return SCSI_MLQUEUE_HOST_BUSY;

dev_not_found:
	sc->result = (DID_NO_CONNECT << 16);
	scsi_set_resid(sc, scsi_bufflen(sc));
	sc->scsi_done(sc);
	return 0;
}



/**
 * bnx2i_iscsi_tmf_timer - iSCSI TMF timeout routine
 *
 * @data: 		pointer to TMF command struct
 *
 * TMF request timeout handler, triggers session recovery process
 */
static void bnx2i_iscsi_tmf_timer(unsigned long data)
{
	struct bnx2i_cmd *cmd = (struct bnx2i_cmd *) data;

	printk(KERN_ALERT "TMF timer: abort failed, cmd 0x%p\n", cmd);
	cmd->cmd_state = ISCSI_CMD_STATE_TMF_TIMEOUT;
	cmd->conn->sess->recovery_state = ISCSI_SESS_RECOVERY_OPEN_ISCSI;
	wake_up(&cmd->conn->sess->er_wait);
	iscsi_conn_error(cmd->conn->cls_conn, ISCSI_ERR_CONN_FAILED);
}


/**
 * bnx2i_initiate_abort_cmd - executes scsi command abort process
 *
 * @sc: 		SCSI-ML command pointer
 *
 * initiate command abort process by requesting CNIC to send
 *	an iSCSI TMF request to target
 */
static int bnx2i_initiate_abort_cmd(struct scsi_cmnd *sc)
{
	struct bnx2i_cmd *cmd = (struct bnx2i_cmd *) sc->SCp.ptr;
	struct bnx2i_cmd *tmf_cmd;
	struct Scsi_Host *shost = cmd->scsi_cmd->device->host;
	struct bnx2i_conn *conn = cmd->conn;
	struct bnx2i_sess *sess;
	struct bnx2i_hba *hba;

	shost = cmd->scsi_cmd->device->host;
	sess = iscsi_hostdata(shost->hostdata);
	BUG_ON(shost != sess->host);

	if (sess && (is_sess_active(sess)))
		hba = sess->hba;
	else
		return FAILED;

	if (test_bit(ADAPTER_STATE_GOING_DOWN, &hba->adapter_state) ||
	    test_bit(ADAPTER_STATE_LINK_DOWN, &hba->adapter_state))
		return FAILED;

	if (cmd->scsi_cmd != sc)
		/* command already completed to scsi mid-layer */
		goto cmd_not_active;

	tmf_cmd = bnx2i_alloc_cmd(sess);
	if (tmf_cmd == NULL)
		goto lack_of_resc;

	sess->tmf_active = 1;
	tmf_cmd->conn = conn = sess->lead_conn;
	tmf_cmd->scsi_cmd = NULL;
	tmf_cmd->iscsi_opcode = ISCSI_OP_SCSI_TMFUNC | ISCSI_OP_IMMEDIATE;
	tmf_cmd->req.cmd_sn = sess->cmdsn;
	tmf_cmd->tmf_ref_itt = cmd->req.itt;
	tmf_cmd->tmf_ref_cmd = cmd;
	tmf_cmd->tmf_ref_sc = cmd->scsi_cmd;
	cmd->cmd_state = ISCSI_CMD_STATE_ABORT_PEND;
	tmf_cmd->cmd_state = ISCSI_CMD_STATE_INITIATED;

#define BNX2I_TMF_TIMEOUT	10 * HZ
	sess->abort_timer.expires = BNX2I_TMF_TIMEOUT + jiffies;
	sess->abort_timer.function = bnx2i_iscsi_tmf_timer;
	sess->abort_timer.data = (unsigned long)tmf_cmd;
	add_timer(&sess->abort_timer);

	bnx2i_send_iscsi_tmf(conn, tmf_cmd);

	/* update iSCSI context for this conn, wait for CNIC to complete */
	wait_event_interruptible(sess->er_wait, (!conn->ep ||
					 (tmf_cmd->cmd_state !=
					  ISCSI_CMD_STATE_TMF_TIMEOUT)));

	if (signal_pending(current))
		flush_signals(current);

	del_timer_sync(&sess->abort_timer);

	if (tmf_cmd->cmd_state == ISCSI_CMD_STATE_TMF_TIMEOUT) {
		printk(KERN_ALERT "abort: abort failed, cmd 0x%p\n", tmf_cmd);
		/* TMF timed out, return error status and let SCSI-ML do
		 * session recovery.
		 */
		list_del_init(&tmf_cmd->link);
		sess->tmf_active = 0;
		bnx2i_free_cmd(sess, tmf_cmd);
		return SUCCESS;
	}

	list_del_init(&tmf_cmd->link);
	sess->tmf_active = 0;
	bnx2i_free_cmd(sess, tmf_cmd);

	if (!cmd->scsi_cmd) {
		goto cmd_completed;
	}
	if ((cmd->scsi_cmd->result & 0xFF0000) == (DID_ABORT << 16)) {
		cmd->cmd_state = ISCSI_CMD_STATE_CLEANUP_PEND;
		bnx2i_send_cmd_cleanup_req(hba, cmd);
		wait_event_interruptible_timeout(sess->er_wait,
						 (cmd->cmd_state ==
						  ISCSI_CMD_STATE_CLEANUP_CMPL),
						 msecs_to_jiffies(
						  ISCSI_CMD_CLEANUP_TIMEOUT));

		if (signal_pending(current))
			flush_signals(current);
	} else
		cmd->scsi_cmd->result = (DID_ABORT << 16);

	list_del_init(&cmd->link);
	bnx2i_return_failed_command(sess, cmd, DID_ABORT);
	bnx2i_free_cmd(sess, cmd);

cmd_completed:
cmd_not_active:
	return SUCCESS;

lack_of_resc:
	return FAILED;
}



static void bnx2i_wait_for_tmf_completion(struct bnx2i_sess *sess)
{
	int lpcnt = 20;
	while (lpcnt-- && sess->tmf_active) {
		msleep(1000);
	}
}


/**
 * bnx2i_abort - 'eh_abort_handler' api function to abort an oustanding
 *			scsi command
 *
 * @sc: 		SCSI-ML command pointer
 *
 * SCSI abort request handler.
 */
int bnx2i_abort(struct scsi_cmnd *sc)
{
	int reason;
	struct bnx2i_cmd *cmd = (struct bnx2i_cmd *) sc->SCp.ptr;

	if (unlikely(!cmd)) {
		/* command already completed to scsi mid-layer
		printk(KERN_INFO "bnx2i_abort: sc 0x%p, not active\n", sc);
		*/
		return SUCCESS;
	}

	reason = bnx2i_initiate_abort_cmd(sc);
	return reason;
}



/**
 * bnx2i_return_failed_command - return failed command back to SCSI-ML
 *
 * @sess:		iscsi session pointer
 * @cmd:		iscsi cmd pointer
 * @err_code: 		SCSI-ML error code, DID_ABORT, DID_BUS_BUSY
 *
 * completes scsi command with appropriate error code to SCSI-ML
 */
void bnx2i_return_failed_command(struct bnx2i_sess *sess,
				 struct bnx2i_cmd *cmd, int err_code)
{
	struct scsi_cmnd *sc = cmd->scsi_cmd;
	if (!sc) {
		return;
	}
	sc->result = err_code << 16;
	scsi_set_resid(sc, scsi_bufflen(cmd->scsi_cmd));
	cmd->scsi_cmd = NULL;
	sess->num_active_cmds--;
	sc->SCp.ptr = NULL;
	sc->scsi_done(sc);
}



/**
 * bnx2i_host_reset - 'eh_host_reset_handler' entry point
 *
 * @sc: 		SCSI-ML command pointer
 *
 * SCSI host reset handler - iSCSI session recovery
 */
int bnx2i_host_reset(struct scsi_cmnd *sc)
{
	struct Scsi_Host *shost;
	struct bnx2i_sess *sess;
	int rc = 0;

	shost = sc->device->host;
	sess = iscsi_hostdata(shost->hostdata);
	printk(KERN_INFO "bnx2i: attempting to reset host, #%d\n",
			  sess->host->host_no);

	BUG_ON(shost != sess->host);
	rc = bnx2i_do_iscsi_sess_recovery(sess, DID_RESET);

	if (!list_empty(&sess->active_cmds))
		bnx2i_flush_active_cmd_queue(sess, DID_NO_CONNECT);

	return rc;
}



/**********************************************************************
 *		open-iscsi interface
 **********************************************************************/


/**
 * bnx2i_register_xport - register a bnx2i device transport name with
 *			the iscsi transport module
 *
 * @hba: 		pointer to adapter instance
 *
 * allocates memory for SCSI host template, iSCSI template and registers
 *	this instance of NX2 device with iSCSI transport kernel module.
 */
int bnx2i_register_xport(struct bnx2i_hba *hba)
{
	void *mem_ptr;
	u32 pci_bus_no;
	u32 pci_dev_no;
	u32 pci_func_no;
	u32 extra;
	struct ethtool_drvinfo drv_info;

	mem_ptr = kmalloc(sizeof(struct scsi_host_template), GFP_KERNEL);
	hba->scsi_template = mem_ptr;
	if (hba->scsi_template == NULL) {
		printk(KERN_ALERT "bnx2i: failed to alloc memory for sht\n");
		return -ENOMEM;
	}

	mem_ptr = kmalloc(sizeof(struct iscsi_transport), GFP_KERNEL);
	hba->iscsi_transport = mem_ptr;
	if (hba->iscsi_transport == NULL) {
		printk(KERN_ALERT "mem error for iscsi_transport template\n");
		goto iscsi_xport_err;
	}

	mem_ptr = kmalloc(BRCM_ISCSI_XPORT_NAME_SIZE_MAX, GFP_KERNEL);
	if (mem_ptr == NULL) {
		printk(KERN_ALERT "failed to alloc memory for xport name\n");
		goto scsi_name_mem_err;
	}

	memcpy(hba->scsi_template, (const void *) &bnx2i_host_template,
	       sizeof(struct scsi_host_template));
	hba->scsi_template->name = mem_ptr;
	memcpy((void *) hba->scsi_template->name,
	       (const void *) bnx2i_host_template.name,
	       strlen(bnx2i_host_template.name) + 1);

	mem_ptr = kmalloc(BRCM_ISCSI_XPORT_NAME_SIZE_MAX, GFP_KERNEL);
	if (mem_ptr == NULL) {
		printk(KERN_ALERT "failed to alloc proc name mem\n");
		goto scsi_proc_name_mem_err;
	}
	hba->scsi_template->proc_name = mem_ptr;
	/* Can't determine device type, 5706/5708 has 40-bit dma addr limit */
	hba->scsi_template->dma_boundary = DMA_40BIT_MASK;
	hba->scsi_template->can_queue = hba->max_sqes -
					BNX2I_RESERVED_SLOW_PATH_CMD_SLOTS;
	hba->scsi_template->cmd_per_lun	= hba->scsi_template->can_queue / 4;

	memcpy((void *) hba->iscsi_transport,
	       (const void *) &bnx2i_iscsi_transport,
	       sizeof(struct iscsi_transport));

	hba->iscsi_transport->host_template = hba->scsi_template;

	mem_ptr = kmalloc(BRCM_ISCSI_XPORT_NAME_SIZE_MAX, GFP_KERNEL);
	if (mem_ptr == NULL) {
		printk(KERN_ALERT "mem alloc error, iscsi xport name\n");
		goto xport_name_mem_err;
	}
	hba->iscsi_transport->name = mem_ptr;

	if (hba->cnic && hba->cnic->netdev) {
		hba->cnic->netdev->ethtool_ops->get_drvinfo(hba->cnic->netdev,
							    &drv_info);
		sscanf(drv_info.bus_info, "%x:%x:%x.%d", &extra,
		       &pci_bus_no, &pci_dev_no, &pci_func_no);

		sprintf(mem_ptr, "%s-%.2x%.2x%.2x", BRCM_ISCSI_XPORT_NAME_PREFIX,
			 (u8)pci_bus_no, (u8)pci_dev_no, (u8)pci_func_no);
	}

	memcpy((void *) hba->scsi_template->proc_name, (const void *) mem_ptr,
	       strlen(mem_ptr) + 1);

	hba->shost_template = iscsi_register_transport(hba->iscsi_transport);
	if (!hba->shost_template) {
		printk(KERN_ALERT "bnx2i: xport reg failed, hba 0x%p\n", hba);
		goto failed_registration;
	}
	printk(KERN_ALERT "bnx2i: netif=%s, iscsi=%s\n",
			  hba->cnic->netdev->name, hba->scsi_template->proc_name);
	return 0;

failed_registration:
	kfree(hba->iscsi_transport->name);
xport_name_mem_err:
	kfree(hba->scsi_template->proc_name);
scsi_proc_name_mem_err:
	kfree(hba->scsi_template->name);
scsi_name_mem_err:
	kfree(hba->iscsi_transport);
iscsi_xport_err:
	kfree(hba->scsi_template);
	printk(KERN_ALERT "register iscsi xport failed, hba 0x%p\n", hba);
	return -ENOMEM;
}


/**
 * bnx2i_deregister_xport - unregisters bnx2i adapter's iscsi transport name
 *
 * @hba: 		pointer to adapter instance
 * 
 * de-allocates memory for SCSI host template, iSCSI template and de-registers
 *	a NX2 device instance
 */
int bnx2i_deregister_xport(struct bnx2i_hba *hba)
{
	iscsi_unregister_transport(hba->iscsi_transport);
	hba->shost_template = NULL;

	kfree(hba->scsi_template->proc_name);
	kfree(hba->scsi_template->name);
	hba->scsi_template->name = NULL;

	kfree(hba->scsi_template);
	hba->scsi_template = NULL;

	kfree(hba->iscsi_transport->name);
	hba->iscsi_transport->name = NULL;

	kfree(hba->iscsi_transport);
	hba->iscsi_transport = NULL;

	return 0;
}


/**
 * bnx2i_session_create - create a new iscsi session
 *
 * @it: 		iscsi transport pointer
 * @scsit: 		scsi transport template pointer
 * @cmds_max: 		max commands supported
 * @qdepth: 		scsi queue depth to support
 * @initial_cmdsn: 	initial iscsi CMDSN to be used for this session
 * @host_no: 		pointer to u32 to return host no
 *
 * Creates a new iSCSI session instance on given device.
 */
struct iscsi_cls_session *
	bnx2i_session_create(struct iscsi_transport *it,
			     struct scsi_transport_template *scsit,
#ifdef _CREATE_SESS_NEW_
			     uint16_t cmds_max, uint16_t qdepth,
#endif
			     uint32_t initial_cmdsn, uint32_t *host_no)
{
	struct bnx2i_hba *hba;
	struct bnx2i_sess *sess;
	struct Scsi_Host *shost;
	struct iscsi_cls_session *cls_session;
	int ret_code;

	hba = bnx2i_get_hba_from_template(scsit);
	if (bnx2i_adapter_ready(hba))
		return NULL;

	shost = scsi_host_alloc(hba->iscsi_transport->host_template,
				hostdata_privsize(sizeof(struct bnx2i_sess)));
	if (!shost)
		return NULL;

	shost->max_id = 1;
	shost->max_channel = 1;
	shost->max_lun = hba->iscsi_transport->max_lun;
	shost->max_cmd_len = hba->iscsi_transport->max_cmd_len;
#ifdef _NEW_CREATE_SESSION_
	if (cmds_max)
		shost->can_queue = cmds_max;
	if (qdepth)
		shost->cmd_per_lun = qdepth;
#endif
	shost->transportt = scsit;
	*host_no = shost->host_no;
	sess = iscsi_hostdata(shost->hostdata);

	if (!sess)
		goto sess_resc_fail;

	memset(sess, 0, sizeof(struct bnx2i_sess));
	sess->hba = hba;
	sess->host = shost;

	/*
	 * For Open-iSCSI, only normal sessions go through bnx2i.
	 * Discovery session goes through host stack TCP/IP stack.
	 */
	ret_code = bnx2i_iscsi_sess_new(hba, sess);
	if (ret_code) {
		/* failed to allocate memory */
		printk(KERN_ALERT "bnx2i_sess_create: unable to alloc sess\n");
		goto sess_resc_fail;
	}

	/* Update CmdSN related parameters */
	sess->cmdsn = initial_cmdsn;
	sess->exp_cmdsn = initial_cmdsn + 1;
	sess->max_cmdsn = initial_cmdsn + 1;

	if (scsi_add_host(shost, NULL))
		goto add_sh_fail;

	if (!try_module_get(it->owner))
		goto cls_sess_falied;

	cls_session = iscsi_create_session(shost, it, 0);
	if (!cls_session)
		goto module_put;
	*(unsigned long *)shost->hostdata = (unsigned long)cls_session;

	return hostdata_session(shost->hostdata);

module_put:
	module_put(it->owner);
cls_sess_falied:
	scsi_remove_host(shost);
add_sh_fail:
	bnx2i_iscsi_sess_release(hba, sess);
sess_resc_fail:
	scsi_host_put(shost);
	return NULL;
}


/**
 * bnx2i_session_destroy - destroys iscsi session
 *
 * @cls_session: 	pointer to iscsi cls session
 *
 * Destroys previously created iSCSI session instance and releases
 *	all resources held by it
 */
void bnx2i_session_destroy(struct iscsi_cls_session *cls_session)
{
	struct Scsi_Host *shost = iscsi_session_to_shost(cls_session);
	struct bnx2i_sess *sess = iscsi_hostdata(shost->hostdata);
	struct module *owner = cls_session->transport->owner;

        iscsi_remove_session(cls_session);
        scsi_remove_host(shost);

	bnx2i_iscsi_sess_release(sess->hba, sess);

	kfree(sess->target_name);
	sess->target_name = NULL;

        iscsi_free_session(cls_session);
	scsi_host_put(shost);
        module_put(owner);
}


/**
 * bnx2i_sess_recovery_timeo - session recovery timeout handler
 *
 * @cls_session: 	pointer to iscsi cls session
 *
 * session recovery timeout handling routine
 */
void bnx2i_sess_recovery_timeo(struct iscsi_cls_session *cls_session)
{
	struct Scsi_Host *shost = iscsi_session_to_shost(cls_session);
	struct bnx2i_sess *sess = iscsi_hostdata(shost->hostdata);

	sess->recovery_state |= ISCSI_SESS_RECOVERY_FAILED;
	wake_up(&sess->er_wait);
}


/**
 * bnx2i_conn_create - create iscsi connection instance
 *
 * @cls_session: 	pointer to iscsi cls session
 * @cid: 		iscsi cid as per rfc (not NX2's CID terminology)
 *
 * Creates a new iSCSI connection instance for a given session
 */
struct iscsi_cls_conn *bnx2i_conn_create(struct iscsi_cls_session *cls_session,
					 uint32_t cid)
{
	struct Scsi_Host *shost = iscsi_session_to_shost(cls_session);
	struct bnx2i_sess *sess = iscsi_hostdata(shost->hostdata);
	struct bnx2i_conn *conn;
	struct iscsi_cls_conn *cls_conn;

	cls_conn = iscsi_create_conn(cls_session, cid);
	if (!cls_conn)
		return NULL;

	conn = cls_conn->dd_data;
	memset(conn, 0, sizeof(struct bnx2i_conn));
	conn->cls_conn = cls_conn;
	conn->exp_statsn = STATSN_UPDATE_SIGNATURE;
	conn->state = CONN_STATE_IDLE;
	/* Initialize the connection structure */
	bnx2i_iscsi_conn_new(sess, conn);
	conn->conn_cid = cid;
	return cls_conn;
}


/**
 * bnx2i_conn_bind - binds iscsi sess, conn and ep objects together
 *
 * @cls_session: 	pointer to iscsi cls session
 * @cls_conn: 		pointer to iscsi cls conn
 * @transport_fd: 	64-bit EP handle
 * @is_leading: 	leading connection on this session?
 *
 * Binds together iSCSI session instance, iSCSI connection instance
 *	and the TCP connection. This routine returns error code if
 *	TCP connection does not belong on the device iSCSI sess/conn
 *	is bound
 */
int bnx2i_conn_bind(struct iscsi_cls_session *cls_session,
		    struct iscsi_cls_conn *cls_conn,
		    uint64_t transport_fd, int is_leading)
{
	struct Scsi_Host *shost = iscsi_session_to_shost(cls_session);
	struct bnx2i_sess *sess = iscsi_hostdata(shost->hostdata);
	struct bnx2i_conn *tmp;
	struct bnx2i_conn *conn = cls_conn->dd_data;
	int ret_code;
	struct bnx2i_endpoint *ep;

	ep = (struct bnx2i_endpoint *) (unsigned long) transport_fd;

	if ((ep->state == EP_STATE_TCP_FIN_RCVD) ||
	    (ep->state == EP_STATE_TCP_RST_RCVD))
		/* Peer disconnect via' FIN or RST */
		return -EINVAL;

	if (ep->hba != sess->hba) {
		/* Error - TCP connection does not belong to this device
		 */
		printk(KERN_ALERT "bnx2i: conn bind, ep=0x%p (%s) does not",
				  ep, ep->hba->netdev->name);
		printk(KERN_ALERT "belong to hba (%s)\n",
				  sess->hba->netdev->name);
		return -EEXIST;
	}
	if (!conn->gen_pdu.cmd)
		conn->gen_pdu.cmd = bnx2i_alloc_cmd(sess);

	/* adjust dma boundary limit which was set to lower bound of 40-bit
	 * address as required by 5706/5708. 5709/57710 does not have any
	 * address limitation requirements. 'dma_mask' parameter is set
	 * by bnx2 module based on device requirements, we just use whatever
	 * is set.
	 */
	shost->dma_boundary = ep->hba->pcidev->dma_mask;

	/* look-up for existing connection, MC/S is not currently supported */
	spin_lock_bh(&sess->lock);
	tmp = NULL;
	if (!list_empty(&sess->conn_list)) {
		list_for_each_entry(tmp, &sess->conn_list, link) {
			if (tmp == conn)
				break;
		}
	}
	if ((tmp != conn) && (conn->sess == sess)) {
		/* bind iSCSI connection to this session */
		list_add(&conn->link, &sess->conn_list);
		if (is_leading)
			sess->lead_conn = conn;
	}

	if (conn->ep) {
		/* This happens when 'iscsid' is killed and restarted. Daemon
		 * has no clue of tranport handle, but knows active conn/sess
		 * and tried to rebind a new tranport (EP) to already active
		 * iSCSI session/connection
		 */
		spin_unlock_bh(&sess->lock);
		bnx2i_ep_disconnect((uint64_t) (unsigned long) conn->ep);
		spin_lock_bh(&sess->lock);
	}

	conn->ep = (struct bnx2i_endpoint *) (unsigned long) transport_fd;
	conn->ep->conn = conn;
	conn->ep->sess = sess;
	conn->state = CONN_STATE_XPORT_READY;
	conn->iscsi_conn_cid = conn->ep->ep_iscsi_cid;
	conn->fw_cid = conn->ep->ep_cid;

	ret_code = bnx2i_bind_conn_to_iscsi_cid(conn, ep->ep_iscsi_cid);
	spin_unlock_bh(&sess->lock);

	/* 5706/5708/5709 FW takes RQ as full when initiated, but for 57710
	 * driver needs to explicitly replenish RQ index during setup.
  	 */
	if (test_bit(BNX2I_NX2_DEV_57710, &ep->hba->cnic_dev_type))
		bnx2i_put_rq_buf(conn, 0);

	bnx2i_arm_cq_event_coalescing(conn->ep, CNIC_ARM_CQE);
	return ret_code;
}


/**
 * bnx2i_conn_destroy - destroy iscsi connection instance & release resources
 *
 * @cls_conn: 		pointer to iscsi cls conn
 *
 * Destroy an iSCSI connection instance and release memory resources held by
 *	this connection
 */
void bnx2i_conn_destroy(struct iscsi_cls_conn *cls_conn)
{
	struct bnx2i_conn *conn = cls_conn->dd_data;
	struct bnx2i_sess *sess = conn->sess;

	bnx2i_conn_free_login_resources(conn->sess->hba, conn);

	list_del_init(&conn->link);
	if (sess->lead_conn == conn)
		sess->lead_conn = NULL;
	if (conn->ep) {
		conn->ep->conn = NULL;
		conn->ep = NULL;
	}

	kfree(conn->persist_address);
	conn->persist_address = NULL;
	iscsi_destroy_conn(cls_conn);
}


/**
 * bnx2i_conn_set_param - set iscsi connection parameter
 *
 * @cls_conn: 		pointer to iscsi cls conn
 * @param: 		parameter type identifier
 * @buf: 		buffer pointer
 * @buflen: 		buffer length
 *
 * During FFP migration, user daemon will issue this call to
 *	update negotiated iSCSI parameters to driver.
 */
int bnx2i_conn_set_param(struct iscsi_cls_conn *cls_conn,
			 enum iscsi_param param, char *buf, int buflen)
{
	struct bnx2i_conn *conn = cls_conn->dd_data;
	struct bnx2i_sess *sess = conn->sess;
	int ret_val = 0;

	spin_lock_bh(&sess->lock);
	if (conn->state != CONN_STATE_IN_LOGIN) {
		printk(KERN_ERR "bnx2i: can't change param [%d]\n", param);
		spin_unlock_bh(&sess->lock);
		return -1;
	}
	spin_unlock_bh(&sess->lock);
	switch (param) {
	case ISCSI_PARAM_MAX_RECV_DLENGTH:
		sscanf(buf, "%d", &conn->max_data_seg_len_recv);
		break;
	case ISCSI_PARAM_MAX_XMIT_DLENGTH:
		sscanf(buf, "%d", &conn->max_data_seg_len_xmit);
		break;
	case ISCSI_PARAM_HDRDGST_EN:
		sscanf(buf, "%d", &conn->header_digest_en);
		break;
	case ISCSI_PARAM_DATADGST_EN:
		sscanf(buf, "%d", &conn->data_digest_en);
		break;
	case ISCSI_PARAM_INITIAL_R2T_EN:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->initial_r2t);
		break;
	case ISCSI_PARAM_MAX_R2T:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->max_r2t);
		break;
	case ISCSI_PARAM_IMM_DATA_EN:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->imm_data);
		break;
	case ISCSI_PARAM_FIRST_BURST:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->first_burst_len);
		break;
	case ISCSI_PARAM_MAX_BURST:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->max_burst_len);
		break;
	case ISCSI_PARAM_PDU_INORDER_EN:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->pdu_inorder);
		break;
	case ISCSI_PARAM_DATASEQ_INORDER_EN:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->dataseq_inorder);
		break;
	case ISCSI_PARAM_ERL:
		if (conn == sess->lead_conn)
			sscanf(buf, "%d", &sess->erl);
		break;
	case ISCSI_PARAM_IFMARKER_EN:
		sscanf(buf, "%d", &conn->ifmarker_enable);
		BUG_ON(conn->ifmarker_enable);
		break;
	case ISCSI_PARAM_OFMARKER_EN:
		sscanf(buf, "%d", &conn->ofmarker_enable);
		BUG_ON(conn->ofmarker_enable);
		break;
	case ISCSI_PARAM_EXP_STATSN:
		sscanf(buf, "%u", &conn->exp_statsn);
		break;
	case ISCSI_PARAM_TARGET_NAME:
		if (sess->target_name)
			break;
		sess->target_name = kstrdup(buf, GFP_KERNEL);
		if (!sess->target_name)
			ret_val = -ENOMEM;
		break;
	case ISCSI_PARAM_TPGT:
		sscanf(buf, "%d", &sess->tgt_prtl_grp);
		break;
	case ISCSI_PARAM_PERSISTENT_PORT:
		sscanf(buf, "%d", &conn->persist_port);
		break;
	case ISCSI_PARAM_PERSISTENT_ADDRESS:
		if (conn->persist_address)
			break;
		conn->persist_address = kstrdup(buf, GFP_KERNEL);
		if (!conn->persist_address)
			ret_val = -ENOMEM;
		break;
	default:
		printk(KERN_ALERT "PARAM_UNKNOWN: 0x%x\n", param);
		ret_val = -ENOSYS;
		break;
	}

	return ret_val;
}


/**
 * bnx2i_conn_get_param - return iscsi connection parameter to caller
 *
 * @cls_conn: 		pointer to iscsi cls conn
 * @param: 		parameter type identifier
 * @buf: 		buffer pointer
 *
 * returns iSCSI connection parameters
 */
int bnx2i_conn_get_param(struct iscsi_cls_conn *cls_conn,
			 enum iscsi_param param, char *buf)
{
	struct bnx2i_conn *conn;
	int len;

	conn = (struct bnx2i_conn *)cls_conn->dd_data;
	if (!conn || !conn->ep ||
	    (conn->ep->state != EP_STATE_ULP_UPDATE_COMPL))
		return -EINVAL;

	len = 0;
	switch (param) {
	case ISCSI_PARAM_MAX_RECV_DLENGTH:
		len = sprintf(buf, "%u\n", conn->max_data_seg_len_recv);
		break;
	case ISCSI_PARAM_MAX_XMIT_DLENGTH:
		len = sprintf(buf, "%u\n", conn->max_data_seg_len_xmit);
		break;
	case ISCSI_PARAM_HDRDGST_EN:
		len = sprintf(buf, "%d\n", conn->header_digest_en);
		break;
	case ISCSI_PARAM_DATADGST_EN:
		len = sprintf(buf, "%d\n", conn->data_digest_en);
		break;
	case ISCSI_PARAM_IFMARKER_EN:
		len = sprintf(buf, "%u\n", conn->ifmarker_enable);
		break;
	case ISCSI_PARAM_OFMARKER_EN:
		len = sprintf(buf, "%u\n", conn->ofmarker_enable);
		break;
	case ISCSI_PARAM_EXP_STATSN:
		len = sprintf(buf, "%u\n", conn->exp_statsn);
		break;
	case ISCSI_PARAM_PERSISTENT_PORT:
		len = sprintf(buf, "%d\n", conn->persist_port);
		break;
	case ISCSI_PARAM_PERSISTENT_ADDRESS:
		if (conn->persist_address)
			len = sprintf(buf, "%s\n", conn->persist_address);
		break;
	case ISCSI_PARAM_CONN_PORT:
		len = sprintf(buf, "%hu\n", conn->ep->cm_sk->dst_port);
		break;
	case ISCSI_PARAM_CONN_ADDRESS:
		len = sprintf(buf, NIPQUAD_FMT "\n",
			      NIPQUAD(conn->ep->cm_sk->dst_ip));
		break;
	default:
		printk(KERN_ALERT "get_param: conn 0x%p param %d not found\n",
				  conn, (u32)param);
		len = -ENOSYS;
	}

	return len;
}


/**
 * bnx2i_session_get_param - returns iscsi session parameter
 *
 * @cls_session: 	pointer to iscsi cls session
 * @param: 		parameter type identifier
 * @buf: 		buffer pointer
 *
 * returns iSCSI session parameters
 */
int bnx2i_session_get_param(struct iscsi_cls_session *cls_session,
			    enum iscsi_param param, char *buf)
{
	struct Scsi_Host *shost;
	struct bnx2i_sess *sess;
	int len;

	shost = iscsi_session_to_shost(cls_session);
	sess = iscsi_hostdata(shost->hostdata);

	len = 0;
	switch (param) {
	case ISCSI_PARAM_INITIAL_R2T_EN:
		len = sprintf(buf, "%d\n", sess->initial_r2t);
		break;
	case ISCSI_PARAM_MAX_R2T:
		len = sprintf(buf, "%hu\n", sess->max_r2t);
		break;
	case ISCSI_PARAM_IMM_DATA_EN:
		len = sprintf(buf, "%d\n", sess->imm_data);
		break;
	case ISCSI_PARAM_FIRST_BURST:
		len = sprintf(buf, "%u\n", sess->first_burst_len);
		break;
	case ISCSI_PARAM_MAX_BURST:
		len = sprintf(buf, "%u\n", sess->max_burst_len);
		break;
	case ISCSI_PARAM_PDU_INORDER_EN:
		len = sprintf(buf, "%d\n", sess->pdu_inorder);
		break;
	case ISCSI_PARAM_DATASEQ_INORDER_EN:
		len = sprintf(buf, "%d\n", sess->dataseq_inorder);
		break;
	case ISCSI_PARAM_ERL:
		len = sprintf(buf, "%d\n", sess->erl);
		break;
	case ISCSI_PARAM_TARGET_NAME:
		if (sess->target_name)
			len = sprintf(buf, "%s\n", sess->target_name);
		break;
	case ISCSI_PARAM_TPGT:
		len = sprintf(buf, "%d\n", sess->tgt_prtl_grp);
		break;
	default:
		printk(KERN_ALERT "sess_get_param: sess 0x%p", sess);
		printk(KERN_ALERT  "param (0x%x) not found\n", (u32) param);
		return -ENOSYS;
	}

	return len;
}


/**
 * bnx2i_conn_start - completes iscsi connection migration to FFP
 *
 * @cls_conn: 		pointer to iscsi cls conn
 *
 * last call in FFP migration to handover iscsi conn to the driver
 */
int bnx2i_conn_start(struct iscsi_cls_conn *cls_conn)
{
	struct bnx2i_conn *conn = (struct bnx2i_conn *) cls_conn->dd_data;
	struct bnx2i_sess *sess;

	if (conn->state != CONN_STATE_IN_LOGIN) {
		printk(KERN_ALERT "conn_start: conn 0x%p state 0x%x err!!\n",
				  conn, conn->state);
		return -EINVAL;
	}
	sess = conn->sess;

	if ((sess->imm_data || !sess->initial_r2t) &&
		sess->first_burst_len > sess->max_burst_len) {
		printk(KERN_ALERT "bnx2i: invalid params, FBL > MBL\n");
			return -EINVAL;
	}

	conn->state = CONN_STATE_FFP_STATE;
	if (sess->lead_conn == conn)
		sess->state = BNX2I_SESS_IN_FFP;

	conn->ep->state = EP_STATE_ULP_UPDATE_START;
	bnx2i_update_iscsi_conn(conn);

	conn->ep->ofld_timer.expires = 10*HZ + jiffies;
	conn->ep->ofld_timer.function = bnx2i_ep_ofld_timer;
	conn->ep->ofld_timer.data = (unsigned long) conn->ep;
	add_timer(&conn->ep->ofld_timer);
	/* update iSCSI context for this conn, wait for CNIC to complete */
	wait_event_interruptible(conn->ep->ofld_wait,
				 conn->ep->state != EP_STATE_ULP_UPDATE_START);

	if (signal_pending(current))
		flush_signals(current);
	del_timer_sync(&conn->ep->ofld_timer);

	/* Free login ITT, not required anymore */
	if (conn->gen_pdu.cmd) {
		list_del_init(&conn->gen_pdu.cmd->link);
		bnx2i_free_cmd(sess, conn->gen_pdu.cmd);
		conn->gen_pdu.cmd = NULL;
	}

	switch (conn->stop_state) {
	case STOP_CONN_RECOVER:
		sess->recovery_state = ISCSI_SESS_RECOVERY_COMPLETE;
		sess->state = BNX2I_SESS_IN_FFP;
		iscsi_unblock_session(session_to_cls(sess));
		wake_up(&sess->er_wait);
		break;
	case STOP_CONN_TERM:
		break;
	default:
		;
	}

	return 0;
}

/**
 * bnx2i_conn_stop - stop any further processing on this connection
 *
 * @cls_conn: 		pointer to iscsi cls conn
 * @flags: 		reason for freezing this connection
 *
 * call to take control of iscsi conn from the driver. Could be called
 *	when login failed, when recovery is to be attempted or during
 *	connection teardown
 */
void bnx2i_conn_stop(struct iscsi_cls_conn *cls_conn, int flag)
{
	struct bnx2i_conn *conn = (struct bnx2i_conn *)cls_conn->dd_data;

	conn->stop_state = flag;
	iscsi_block_session(session_to_cls(conn->sess));

	/* Free unfinished login ITT, logout, not required anymore */
	if (conn->gen_pdu.cmd) {
		list_del_init(&conn->gen_pdu.cmd->link);
		bnx2i_free_cmd(conn->sess, conn->gen_pdu.cmd);
		conn->gen_pdu.cmd = NULL;
	}

	switch (flag) {
	case STOP_CONN_RECOVER:
		conn->sess->state = BNX2I_SESS_IN_RECOVERY;
		if (!conn->sess->recovery_state) {	/* nopout timeout */
			unsigned long flags;
			struct Scsi_Host *host = conn->sess->host;
			spin_lock_irqsave(host->host_lock, flags);
			conn->sess->recovery_state =
				ISCSI_SESS_RECOVERY_OPEN_ISCSI;
			spin_unlock_irqrestore(host->host_lock, flags);
		}
		break;
	case STOP_CONN_TERM:
		if (conn->sess && (conn->sess->state & BNX2I_SESS_IN_FFP)) {
			conn->sess->state = BNX2I_SESS_IN_SHUTDOWN;
		}
		break;
	default:
		printk(KERN_ERR "bnx2i: invalid conn stop req %d\n", flag);
	}

	return;
}


/**
 * bnx2i_conn_send_pdu - iscsi transport callback entry point to send
 *			iscsi slow path pdus, such as LOGIN/LOGOUT/NOPOUT, etc
 *
 * @hba: 		pointer to adapter instance
 *
 * sends iSCSI PDUs prepared by user daemon, only login, logout, nop-out pdu
 *	will flow this path.
 */
int bnx2i_conn_send_pdu(struct iscsi_cls_conn *cls_conn,
			struct iscsi_hdr *hdr, char *data,
			uint32_t data_size)
{
	struct bnx2i_conn *conn;
	struct bnx2i_cmd *cmnd;
	uint32_t payload_size;
	int rc;
	unsigned long flags;

	if (!cls_conn) {
		printk(KERN_ALERT "bnx2i_conn_send_pdu: NULL conn ptr. \n");
		return -EIO;
	}
	conn = (struct bnx2i_conn *)cls_conn->dd_data;
	if (!conn->gen_pdu.req_buf) {
		printk(KERN_ALERT "send_pdu: login buf not allocated\n");
		/* ERR - buffer not allocated, should not happen */
		return -EIO;
	}

	if (conn->gen_pdu.cmd) {
		if ((conn->state != CONN_STATE_XPORT_READY) &&
		    (conn->state != CONN_STATE_IN_LOGIN)) {
			printk(KERN_ALERT "send_pdu: %d != XPORT_READY\n",
					  conn->state);
			return -EPERM;
		}
		cmnd = conn->gen_pdu.cmd;
	} else {	/* could be NOPOUT or the LOGOUT request */
		spin_lock_irqsave(conn->sess->host->host_lock, flags);
		cmnd = bnx2i_alloc_cmd(conn->sess);
		spin_unlock_irqrestore(conn->sess->host->host_lock, flags);

		if (!cmnd) {
			printk(KERN_ALERT "bnx2i: Error - cmd not allocated\n");
			return -EIO;
		}
	}
	memset(conn->gen_pdu.req_buf, 0, ISCSI_CONN_LOGIN_BUF_SIZE);
	/* Login request, copy hdr & data to buffer in conn struct */
	memcpy(&conn->gen_pdu.pdu_hdr, (const void *) hdr,
	       sizeof(struct iscsi_hdr));

	cmnd->iscsi_opcode = hdr->opcode;
	switch (hdr->opcode & ISCSI_OPCODE_MASK) {
	case ISCSI_OP_LOGIN:
		if (conn->state == CONN_STATE_XPORT_READY)
			conn->state = CONN_STATE_IN_LOGIN;
		break;
	case ISCSI_OP_LOGOUT:
		if (!conn->gen_pdu.cmd)
			conn->gen_pdu.cmd = cmnd;
		conn->state = CONN_STATE_IN_LOGOUT;
		conn->sess->state = BNX2I_SESS_IN_LOGOUT;
		if (conn->sess->tmf_active)
			bnx2i_wait_for_tmf_completion(conn->sess);
		break;
	case ISCSI_OP_NOOP_OUT:
		break;
	default:
		;
	}

	conn->gen_pdu.req_buf_size = data_size;
	payload_size = (hdr->dlength[0] << 16) | (hdr->dlength[1] << 8) |
		       hdr->dlength[2];

	if (data_size) {
		memcpy(conn->gen_pdu.req_buf, (const void *)data, data_size);
		conn->gen_pdu.req_wr_ptr =
			conn->gen_pdu.req_buf + payload_size;
	}
	cmnd->conn = conn;
	cmnd->scsi_cmd = NULL;
	rc = bnx2i_iscsi_send_generic_request(cmnd);
	return rc;
}


/**
 * bnx2i_conn_get_stats - returns iSCSI stats
 *
 * @cls_conn: 		pointer to iscsi cls conn
 * @stats: 		pointer to iscsi statistic struct
 */
void bnx2i_conn_get_stats(struct iscsi_cls_conn *cls_conn,
			  struct iscsi_stats *stats)
{
	struct bnx2i_conn *conn = (struct bnx2i_conn *) cls_conn->dd_data;

	stats->txdata_octets = conn->total_data_octets_sent;
	stats->rxdata_octets = conn->total_data_octets_rcvd;

	stats->noptx_pdus = conn->num_nopin_pdus;
	stats->scsicmd_pdus = conn->num_scsi_cmd_pdus;
	stats->tmfcmd_pdus = conn->num_tmf_req_pdus;
	stats->login_pdus = conn->num_login_req_pdus;
	stats->text_pdus = 0;
	stats->dataout_pdus = conn->num_dataout_pdus;
	stats->logout_pdus = conn->num_logout_req_pdus;
	stats->snack_pdus = 0;

	stats->noprx_pdus = conn->num_nopout_pdus;
	stats->scsirsp_pdus = conn->num_scsi_resp_pdus;
	stats->tmfrsp_pdus = conn->num_tmf_resp_pdus;
	stats->textrsp_pdus = 0;
	stats->datain_pdus = conn->num_datain_pdus;
	stats->logoutrsp_pdus = conn->num_logout_resp_pdus;
	stats->r2t_pdus = conn->num_r2t_pdus;
	stats->async_pdus = conn->num_async_pdus;
	stats->rjt_pdus = conn->num_reject_pdus;

	stats->digest_err = 0;
	stats->timeout_err = 0;
	stats->custom_length = 0;
}



/**
 * bnx2i_check_nx2_dev_busy - this routine unregister devices if
 *			there are no active conns
 */
void bnx2i_check_nx2_dev_busy(void)
{
	bnx2i_unreg_dev_all();
}


/**
 * bnx2i_check_route - checks if target IP route belongs to one of
 *			NX2 devices
 *
 * @dst_addr: 		target IP address
 *
 * check if route resolves to BNX2 device
 */
static struct bnx2i_hba *bnx2i_check_route(struct sockaddr *dst_addr)
{
	struct sockaddr_in *desti = (struct sockaddr_in *) dst_addr;
	struct bnx2i_hba *hba;
	struct cnic_dev *cnic = NULL;

	bnx2i_reg_dev_all();

	hba = get_adapter_list_head();
	if (hba && hba->cnic)
		cnic = hba->cnic->cm_select_dev(desti, CNIC_ULP_ISCSI);

	if (!cnic) {
		printk(KERN_ALERT "bnx2i: check route, can't connect using cnic\n");
		goto no_nx2_route;
	}
	hba = bnx2i_find_hba_for_cnic(cnic);
	if (!hba) {
		goto no_nx2_route;
	}

	if (bnx2i_adapter_ready(hba)) {
		printk(KERN_ALERT "bnx2i: check route, hba not found\n");
		goto no_nx2_route;
	}
	if (hba->netdev->mtu > hba->mtu_supported) {
		printk(KERN_ALERT "bnx2i: %s network i/f mtu is set to %d\n",
				  hba->netdev->name, hba->netdev->mtu);
		printk(KERN_ALERT "bnx2i: iSCSI HBA can support mtu of %d\n",
				  hba->mtu_supported);
		goto no_nx2_route;
	}
	return hba;
no_nx2_route:
	return NULL;
}


/**
 * bnx2i_tear_down_conn - tear down iscsi/tcp connection and free resources
 *
 * @hba: 		pointer to adapter instance
 * @ep: 		endpoint (transport indentifier) structure
 *
 * destroys cm_sock structure and on chip iscsi context
 */
static int bnx2i_tear_down_conn(struct bnx2i_hba *hba,
				 struct bnx2i_endpoint *ep)
{
	if (test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic))
		hba->cnic->cm_destroy(ep->cm_sk);

	if (test_bit(ADAPTER_STATE_GOING_DOWN, &ep->hba->adapter_state))
		ep->state = EP_STATE_DISCONN_COMPL;

	if (test_bit(BNX2I_NX2_DEV_57710, &hba->cnic_dev_type) &&
	    ep->state == EP_STATE_DISCONN_TIMEDOUT) {
		printk(KERN_ALERT "bnx2i - ERROR - please submit GRC Dump,"
				  " NW/PCIe trace, driver msgs to developers"
				  " for analysis\n");
		return 1;
	}

	ep->state = EP_STATE_CLEANUP_START;
	init_timer(&ep->ofld_timer);
	ep->ofld_timer.expires = 10*HZ + jiffies;
	ep->ofld_timer.function = bnx2i_ep_ofld_timer;
	ep->ofld_timer.data = (unsigned long) ep;
	add_timer(&ep->ofld_timer);

	bnx2i_ep_destroy_list_add(hba, ep);

	/* destroy iSCSI context, wait for it to complete */
	bnx2i_send_conn_destroy(hba, ep);
	wait_event_interruptible(ep->ofld_wait,
				 (ep->state != EP_STATE_CLEANUP_START));

	if (signal_pending(current))
		flush_signals(current);
	del_timer_sync(&ep->ofld_timer);
	bnx2i_ep_destroy_list_del(hba, ep);

	if (ep->state != EP_STATE_CLEANUP_CMPL)
		/* should never happen */
		printk(KERN_ALERT "bnx2i - conn destroy failed\n");
	return 0;
}


/**
 * bnx2i_ep_connect - establish TCP connection to target portal
 *
 * @dst_addr: 		target IP address
 * @non_blocking: 	blocking or non-blocking call
 * @ep_handle: 		placeholder to return new created  endpoint handle
 *
 * this routine initiates the TCP/IP connection by invoking Option-2 i/f
 *	with l5_core and the CNIC. This is a multi-step process of resolving
 *	route to target, create a iscsi connection context, handshaking with
 *	CNIC module to create/initialize the socket struct and finally
 *	sending down option-2 request to complete TCP 3-way handshake
 */
int bnx2i_ep_connect(struct sockaddr *dst_addr, int non_blocking,
		     uint64_t *ep_handle)
{
	u32 iscsi_cid = BNX2I_CID_RESERVED;
	struct sockaddr_in *desti;
	struct sockaddr_in6 *desti6;
	struct bnx2i_endpoint *endpoint;
	struct bnx2i_hba *hba;
	struct cnic_dev *cnic;
	struct cnic_sockaddr saddr;
	int rc = 0;

	/*
	 * check if the given destination can be reached through NX2 device
	 */

	hba = bnx2i_check_route(dst_addr);
	if (!hba) {
		rc = -ENOMEM;
		goto check_busy;
	}

	cnic = hba->cnic;
	endpoint = bnx2i_alloc_ep(hba);
	if (!endpoint) {
		*ep_handle = (uint64_t) 0;
		rc = -ENOMEM;
		goto check_busy;
	}

	spin_lock(&hba->net_dev_lock);

	if (bnx2i_adapter_ready(hba)) {
		rc = -EPERM;
		goto net_if_down;
	}

	endpoint->state = EP_STATE_IDLE;
	endpoint->teardown_mode = BNX2I_ABORTIVE_SHUTDOWN;
	endpoint->ep_iscsi_cid = (u16)ISCSI_RESERVED_TAG;
	iscsi_cid = bnx2i_alloc_iscsi_cid(hba);
	if (iscsi_cid == ISCSI_RESERVED_TAG) {
		printk(KERN_ALERT "alloc_ep: unable to allocate iscsi cid\n");
		rc = -ENOMEM;
		goto iscsi_cid_err;
	}
	endpoint->hba_age = hba->age;

	rc = bnx2i_alloc_qp_resc(hba, endpoint);
	if (rc != 0) {
		printk(KERN_ALERT "bnx2i: ep_conn, alloc QP resc error\n");
		rc = -ENOMEM;
		goto qp_resc_err;
	}

	endpoint->ep_iscsi_cid = iscsi_cid & 0xFFFF;
	endpoint->state = EP_STATE_OFLD_START;
	bnx2i_ep_ofld_list_add(hba, endpoint);

	init_timer(&endpoint->ofld_timer);
	endpoint->ofld_timer.expires = 2 * HZ + jiffies;
	endpoint->ofld_timer.function = bnx2i_ep_ofld_timer;
	endpoint->ofld_timer.data = (unsigned long) endpoint;
	add_timer(&endpoint->ofld_timer);

	bnx2i_send_conn_ofld_req(hba, endpoint);

	/* Wait for CNIC hardware to setup conn context and return 'cid' */
	wait_event_interruptible(endpoint->ofld_wait,
				 endpoint->state != EP_STATE_OFLD_START);

	if (signal_pending(current))
		flush_signals(current);
	del_timer_sync(&endpoint->ofld_timer);
	bnx2i_ep_ofld_list_del(hba, endpoint);
	
	if (endpoint->state != EP_STATE_OFLD_COMPL) {
		rc = -ENOSPC;
		goto conn_failed;
	}

	if (!test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
		rc = -EINVAL;
		goto conn_failed;
	} else
		rc = cnic->cm_create(cnic, CNIC_ULP_ISCSI, endpoint->ep_cid,
				     iscsi_cid, &endpoint->cm_sk, endpoint);
	if (rc) {
		rc = -EINVAL;
		goto conn_failed;
	}

	endpoint->cm_sk->rcv_buf = 256 * 1024;
	endpoint->cm_sk->snd_buf = 256 * 1024;
	clear_bit(SK_TCP_TIMESTAMP, &endpoint->cm_sk->tcp_flags);

	memset(&saddr, 0, sizeof(saddr));

	if (dst_addr->sa_family == AF_INET) {
		desti = (struct sockaddr_in *) dst_addr;
		saddr.remote.v4 = *desti;
		saddr.local.v4.sin_port = htons(endpoint->tcp_port);
		saddr.local.v4.sin_family = desti->sin_family;
	} else if (dst_addr->sa_family == AF_INET6) {
		desti6 = (struct sockaddr_in6 *) dst_addr;
		saddr.remote.v6 = *desti6;
		saddr.local.v6.sin6_port = htons(endpoint->tcp_port);
		saddr.local.v6.sin6_family = desti6->sin6_family;
	}

	endpoint->timestamp = jiffies;
	endpoint->state = EP_STATE_CONNECT_START;
	if (!test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
		rc = -EINVAL;
		goto conn_failed;
	} else
		rc = cnic->cm_connect(endpoint->cm_sk, &saddr);

	if (rc)
		goto release_ep;

	bnx2i_map_ep_dbell_regs(endpoint);
	*ep_handle = (uint64_t) (unsigned long) endpoint;
	spin_unlock(&hba->net_dev_lock);
	return 0;

release_ep:
	bnx2i_tear_down_conn(hba, endpoint);
conn_failed:
net_if_down:
iscsi_cid_err:
	bnx2i_free_qp_resc(hba, endpoint);
qp_resc_err:
	bnx2i_free_ep(endpoint);
	spin_unlock(&hba->net_dev_lock);
check_busy:
	*ep_handle = (uint64_t) 0;
	bnx2i_check_nx2_dev_busy();
	return rc;
}


/**
 * bnx2i_ep_poll - polls for TCP connection establishement
 *
 * @ep_handle: 		TCP connection (endpoint) handle
 * @timeout_ms: 	timeout value in milli secs
 *
 * polls for TCP connect request to complete
 */
int bnx2i_ep_poll(uint64_t ep_handle, int timeout_ms)
{
	struct bnx2i_endpoint *ep;
	int rc = 0;

	ep = (struct bnx2i_endpoint *) (unsigned long) ep_handle;
	if (!ep)
		return -EINVAL;

	if ((ep->state == EP_STATE_IDLE) ||
	    (ep->state == EP_STATE_CONNECT_FAILED) ||
	    (ep->state == EP_STATE_OFLD_FAILED))
		return -1;

	if (ep->state == EP_STATE_CONNECT_COMPL)
		return 1;

	rc = wait_event_interruptible_timeout(ep->ofld_wait,
					      ((ep->state ==
					        EP_STATE_OFLD_FAILED) ||
					       (ep->state ==
					         EP_STATE_CONNECT_FAILED) ||
					       (ep->state ==
					        EP_STATE_CONNECT_COMPL)),
					       msecs_to_jiffies(timeout_ms));
	if (ep->state == EP_STATE_OFLD_FAILED)
		rc = -1;

	if (rc > 0)
		return 1;
	else if (!rc)
		return 0;	/* timeout */
	else
		return rc;
}

/**
 * bnx2i_ep_tcp_conn_active - check EP state transition to check
 *		if underlying TCP connection is active
 *
 * @ep: 		endpoint pointer
 *
 */
static int bnx2i_ep_tcp_conn_active(struct bnx2i_endpoint *ep)
{
	int ret;

	switch (ep->state) {
	case EP_STATE_CONNECT_START:
	case EP_STATE_CLEANUP_FAILED:
	case EP_STATE_OFLD_FAILED:
	case EP_STATE_DISCONN_TIMEDOUT:
		ret = 0;
		break;
	case EP_STATE_CONNECT_COMPL:
	case EP_STATE_ULP_UPDATE_START:
	case EP_STATE_ULP_UPDATE_COMPL:
	case EP_STATE_TCP_FIN_RCVD:
	case EP_STATE_ULP_UPDATE_FAILED:
		ret = 1;
		break;
	case EP_STATE_TCP_RST_RCVD:
		ret = 0;
		break;
	case EP_STATE_CONNECT_FAILED:
		if (test_bit(BNX2I_NX2_DEV_57710, &ep->hba->cnic_dev_type))
			ret = 1;
		else
			ret = 0;
		break;
	default:
		ret = 0;
	}

	return ret;
}

/**
 * bnx2i_ep_disconnect - executes TCP connection teardown process
 *
 * @ep_handle: 		TCP connection (endpoint) handle
 *
 * executes  TCP connection teardown process
 */
void bnx2i_ep_disconnect(uint64_t ep_handle)
{
	struct bnx2i_endpoint *ep;
	struct cnic_dev *cnic;
	struct bnx2i_hba *hba;
	struct bnx2i_sess *sess;

	ep = (struct bnx2i_endpoint *) (unsigned long) ep_handle;
	if (!ep || (ep_handle == -1))
		return;

        while ((ep->state == EP_STATE_CONNECT_START) &&
		!time_after(jiffies, ep->timestamp + (12 * HZ)))
                msleep(250);

	hba = ep->hba;
	if (ep->state == EP_STATE_IDLE)
		goto return_ep;
	cnic = hba->cnic;

	spin_lock(&hba->net_dev_lock);
	if (!test_bit(ADAPTER_STATE_UP, &hba->adapter_state))
		goto free_resc;
	if (ep->hba_age != hba->age)
		goto dev_reset;

	if (!bnx2i_ep_tcp_conn_active(ep))
		goto destory_conn;

	ep->state = EP_STATE_DISCONN_START;

	init_timer(&ep->ofld_timer);
	ep->ofld_timer.expires = 20 * HZ + jiffies;
	ep->ofld_timer.function = bnx2i_ep_ofld_timer;
	ep->ofld_timer.data = (unsigned long) ep;
	add_timer(&ep->ofld_timer);

	if (test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
		if (ep->teardown_mode == BNX2I_GRACEFUL_SHUTDOWN)
			cnic->cm_close(ep->cm_sk);
		else
			cnic->cm_abort(ep->cm_sk);
	} else
		goto free_resc;

	/* wait for option-2 conn teardown */
	wait_event_interruptible(ep->ofld_wait,
				 ep->state != EP_STATE_DISCONN_START);

	if (signal_pending(current))
		flush_signals(current);
	del_timer_sync(&ep->ofld_timer);

destory_conn:
	if (bnx2i_tear_down_conn(hba, ep)) {
		spin_unlock(&hba->net_dev_lock);
		return;
	}

dev_reset:
	/* in case of 3-way handshake failure, there won't be any binding
	 * between EP and SESS
	 */
	if (ep->sess)
		bnx2i_flush_active_cmd_queue(ep->sess, DID_BUS_BUSY);

free_resc:
	spin_unlock(&hba->net_dev_lock);
	bnx2i_free_qp_resc(hba, ep);
return_ep:
	/* check if session recovery in progress */
	sess = ep->sess;
	bnx2i_free_ep(ep);
	if (sess) {
		sess->state = BNX2I_SESS_INITIAL;
		wake_up(&sess->er_wait);
	}
	wake_up_interruptible(&hba->eh_wait);

	if (!hba->ofld_conns_active)
		bnx2i_check_nx2_dev_busy();

	return;
}


int bnx2i_check_ioctl_signature(struct bnx2i_ioctl_header *ioc_hdr)
{
	if (strcmp(ioc_hdr->signature, BNX2I_MGMT_SIGNATURE))
		return -EPERM;
	return 0;
}

static int bnx2i_tcp_port_count_ioctl(struct file *file, unsigned long arg)
{
	struct bnx2i_get_port_count __user *user_ioc =
		(struct bnx2i_get_port_count __user *)arg;
	struct bnx2i_get_port_count ioc_req;
	int error = 0;
	unsigned int count = 0;

	if (copy_from_user(&ioc_req, user_ioc, sizeof(ioc_req))) {
		error = -EFAULT;
		goto out;
	}

	error = bnx2i_check_ioctl_signature(&ioc_req.hdr);
	if (error)
		goto out;

	if (bnx2i_tcp_port_tbl.num_free_ports < 10 &&
	    bnx2i_tcp_port_tbl.num_required) {
		if (bnx2i_tcp_port_tbl.num_required < 32)
			count = bnx2i_tcp_port_tbl.num_required;
		else
			count = 32;
	}

	ioc_req.port_count = count;

	if (copy_to_user(&user_ioc->port_count, &ioc_req.port_count,
			 sizeof(ioc_req.port_count))) {
		error = -EFAULT;
		goto out;
	}

out:
	return error;
}


static int bnx2i_tcp_port_ioctl(struct file *file, unsigned long arg)
{
	struct bnx2i_set_port_num __user *user_ioc =
		(struct bnx2i_set_port_num __user *)arg;
	struct bnx2i_set_port_num ioc_req;
	struct bnx2i_set_port_num *ioc_req_mp = NULL;
	int ioc_msg_size = sizeof(ioc_req);
	int error;
	int i;

	if (copy_from_user(&ioc_req, user_ioc, ioc_msg_size)) {
		error = -EFAULT;
		goto out;
	}

	error = bnx2i_check_ioctl_signature(&ioc_req.hdr);
	if (error)
		goto out;

	if (ioc_req.num_ports > 1) {
		ioc_msg_size += (ioc_req.num_ports - 1) *
				sizeof(ioc_req.tcp_port[0]);

		ioc_req_mp = kmalloc(ioc_msg_size, GFP_KERNEL);
		if (!ioc_req_mp)
			goto out;

		if (copy_from_user(ioc_req_mp, user_ioc, ioc_msg_size)) {
			error = -EFAULT;
			goto out_kfree;
		}
	}

	if (ioc_req.num_ports)
		bnx2i_tcp_port_new_entry(ioc_req.tcp_port[0]);

	i = 1;
	while (i < ioc_req_mp->num_ports)
		bnx2i_tcp_port_new_entry(ioc_req_mp->tcp_port[i++]);

	return 0;

out_kfree:
	kfree(ioc_req_mp);
out:
	return error;
}


/*
 * bnx2i_ioctl_init: initialization routine, registers char driver
 */
int bnx2i_ioctl_init(void)
{
	int ret;

        /* Register char device node */
        ret = register_chrdev(0, "bnx2i", &bnx2i_mgmt_fops);

        if (ret < 0) {
                printk(KERN_ERR "bnx2i: failed to register device node\n");
                return ret;
        }

        bnx2i_major_no = ret;

	return 0;
}

void bnx2i_ioctl_cleanup(void)
{
	if (bnx2i_major_no) {
		unregister_chrdev(bnx2i_major_no, "bnx2i");
	}
}

/*
 * bnx2i_mgmt_open -  "open" entry point
 */
static int bnx2i_mgmt_open(struct inode *inode, struct file *filep)
{
        /* only allow access to admin user */
        if (!capable(CAP_SYS_ADMIN)) {
                return -EACCES;
	}

        return 0;
}

/*
 * bnx2i_mgmt_release- "release" entry point
 */
static int bnx2i_mgmt_release(struct inode *inode, struct file *filep)
{
        return 0;
}



/*
 * bnx2i_mgmt_ioctl - char driver ioctl entry point
 */
static int bnx2i_mgmt_ioctl(struct inode *node, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
	long rc = 0;
	switch (cmd) {
		case BNX2I_IOCTL_GET_PORT_REQ:
			rc = bnx2i_tcp_port_count_ioctl(file, arg);
			break;
		case BNX2I_IOCTL_SET_TCP_PORT:
			rc = bnx2i_tcp_port_ioctl(file, arg);
			break;
		default:
			printk(KERN_ERR "bnx2i: unknown ioctl cmd %x\n", cmd);
			return -ENOTTY;
	}

	return rc;
}


#ifdef CONFIG_COMPAT

static int bnx2i_tcp_port_count_compat_ioctl(struct file *file, unsigned long arg)
{
	struct bnx2i_get_port_count __user *user_ioc =
		(struct bnx2i_get_port_count __user *)arg;
	struct bnx2i_get_port_count *ioc_req =
		compat_alloc_user_space(sizeof(struct bnx2i_get_port_count));
	int error;
	unsigned int count = 0;

	if (clear_user(ioc_req, sizeof(*ioc_req)))
		return -EFAULT;

	if (copy_in_user(ioc_req, user_ioc, sizeof(*ioc_req))) {
		error = -EFAULT;
		goto out;
	}

	error = bnx2i_check_ioctl_signature(&ioc_req->hdr);
	if (error)
		goto out;

	if (bnx2i_tcp_port_tbl.num_free_ports < 10 &&
	    bnx2i_tcp_port_tbl.num_required) {
		if (bnx2i_tcp_port_tbl.num_required < 32)
			count = bnx2i_tcp_port_tbl.num_required;
		else
			count = 32;
	}

	if (copy_to_user(&ioc_req->port_count, &count,
			 sizeof(ioc_req->port_count))) {
		error = -EFAULT;
		goto out;
	}

	if (copy_in_user(&user_ioc->port_count, &ioc_req->port_count,
			 sizeof(u32))) {
		error = -EFAULT;
		goto out;
	}
	return 0;

out:
	return error;
}

static int bnx2i_tcp_port_compat_ioctl(struct file *file, unsigned long arg)
{
	struct bnx2i_set_port_num __user *user_ioc =
		(struct bnx2i_set_port_num __user *)arg;
	struct bnx2i_set_port_num *ioc_req =
		compat_alloc_user_space(sizeof(struct bnx2i_set_port_num));
	struct bnx2i_set_port_num *ioc_req_mp = NULL;
	int ioc_msg_size = sizeof(*ioc_req);
	int error;
	int i;

	if (clear_user(ioc_req, sizeof(*ioc_req)))
		return -EFAULT;

	if (copy_in_user(ioc_req, user_ioc, ioc_msg_size)) {
		error = -EFAULT;
		goto out;
	}

	error = bnx2i_check_ioctl_signature(&ioc_req->hdr);
	if (error)
		goto out;

	if (ioc_req->num_ports > 1) {
		ioc_msg_size += (ioc_req->num_ports - 1) *
				sizeof(ioc_req->tcp_port[0]);

		ioc_req_mp = compat_alloc_user_space(ioc_msg_size);
		if (!ioc_req_mp)
			goto out;

		if (copy_in_user(ioc_req_mp, user_ioc, ioc_msg_size)) {
			error = -EFAULT;
			goto out;
		}

		i = 0;
		while ((i < ioc_req_mp->num_ports) && ioc_req_mp)
			bnx2i_tcp_port_new_entry(ioc_req_mp->tcp_port[i++]);

	} else if (ioc_req->num_ports == 1)
		bnx2i_tcp_port_new_entry(ioc_req->tcp_port[0]);

out:
	return error;


}


/*
 * bnx2i_mgmt_compat_ioctl - char node ioctl entry point
 */
static long bnx2i_mgmt_compat_ioctl(struct file *file,
				    unsigned int cmd, unsigned long arg)
{
	int rc = -ENOTTY;

	switch (cmd) {
		case BNX2I_IOCTL_GET_PORT_REQ:
			rc = bnx2i_tcp_port_count_compat_ioctl(file, arg);
			break;
		case BNX2I_IOCTL_SET_TCP_PORT:
			rc = bnx2i_tcp_port_compat_ioctl(file, arg);
			break;
	}

        return rc;
}

#endif

/*
 * File operations structure - management interface
 */
struct file_operations bnx2i_mgmt_fops = {
        .owner = THIS_MODULE,
        .open = bnx2i_mgmt_open,
        .release = bnx2i_mgmt_release,
        .ioctl = bnx2i_mgmt_ioctl,
#ifdef CONFIG_COMPAT
        .compat_ioctl = bnx2i_mgmt_compat_ioctl,
#endif
};


/*
 * 'Scsi_Host_Template' structure and 'iscsi_tranport' structure template
 * used while registering with the iSCSI trnaport module.
 */
struct scsi_host_template bnx2i_host_template = {
	.module				= THIS_MODULE,
	.name				= "Broadcom Offload iSCSI Initiator",
	.queuecommand			= bnx2i_queuecommand,
	.eh_abort_handler		= bnx2i_abort,
	.eh_host_reset_handler		= bnx2i_host_reset,
	.bios_param			= NULL,
	.can_queue			= 128,
	.max_sectors			= 127,
	.this_id			= -1,
	.cmd_per_lun			= 64,
	.use_clustering			= ENABLE_CLUSTERING,
	.sg_tablesize			= ISCSI_MAX_BDS_PER_CMD,
	.proc_name			= NULL
	};



struct iscsi_transport bnx2i_iscsi_transport = {
	.owner			= THIS_MODULE,
	.name			= "bnx2i",
	.caps			= CAP_RECOVERY_L0 | CAP_HDRDGST | CAP_MULTI_R2T
				  | CAP_DATADGST,
	.param_mask		= ISCSI_MAX_RECV_DLENGTH |
				  ISCSI_MAX_XMIT_DLENGTH |
				  ISCSI_HDRDGST_EN |
				  ISCSI_DATADGST_EN |
				  ISCSI_INITIAL_R2T_EN |
				  ISCSI_MAX_R2T |
				  ISCSI_IMM_DATA_EN |
				  ISCSI_FIRST_BURST |
				  ISCSI_MAX_BURST |
				  ISCSI_PDU_INORDER_EN |
				  ISCSI_DATASEQ_INORDER_EN |
				  ISCSI_ERL |
				  ISCSI_CONN_PORT |
				  ISCSI_CONN_ADDRESS |
				  ISCSI_EXP_STATSN |
				  ISCSI_PERSISTENT_PORT |
				  ISCSI_PERSISTENT_ADDRESS |
				  ISCSI_TARGET_NAME |
				  ISCSI_TPGT,
	.host_template		= &bnx2i_host_template,
	.sessiondata_size	= sizeof(struct bnx2i_sess),
	.conndata_size		= sizeof(struct bnx2i_conn),
	.max_conn		= 1,
	.max_cmd_len		= 16,
	.max_lun		= 512,
	.create_session		= bnx2i_session_create,
	.destroy_session	= bnx2i_session_destroy,
	.create_conn		= bnx2i_conn_create,
	.bind_conn		= bnx2i_conn_bind,
	.destroy_conn		= bnx2i_conn_destroy,
	.set_param		= bnx2i_conn_set_param,
	.get_conn_param		= bnx2i_conn_get_param,
	.get_session_param	= bnx2i_session_get_param,
	.start_conn		= bnx2i_conn_start,
	.stop_conn		= bnx2i_conn_stop,
	.send_pdu		= bnx2i_conn_send_pdu,
	.get_stats		= bnx2i_conn_get_stats,
	/* TCP connect - disconnect - option-2 interface calls */
	.ep_connect		= bnx2i_ep_connect,
	.ep_poll		= bnx2i_ep_poll,
	.ep_disconnect		= bnx2i_ep_disconnect,
	/* Error recovery timeout call */
	.session_recovery_timedout = bnx2i_sess_recovery_timeo
};
