/* bnx2i_sysfs.c: Broadcom NetXtreme II iSCSI driver.
 *
 * Copyright (c) 2006 - 2009 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: Anil Veerabhadrappa (anilgv@broadcom.com)
 */

#include "bnx2i.h"
#include <linux/ethtool.h>

#ifdef _SYSFS_INCL_

#define BNX2I_SYSFS_VERSION	0x2

#define BNX2I_PCICFG_REG_WINDOW_ADDRESS		0x00000078
#define BNX2I_PCICFG_REG_WINDOW_ADDRESS_VAL	(0xfffffL<<2)
#define BNX2I_PCICFG_REG_WINDOW			0x00000080

#define BNX2I_CP_IDLE_COUNTER		(BNX2_CP_SCRATCH + 0x28 + 0x10)
#define BNX2I_COM_IDLE_COUNTER		(BNX2_COM_SCRATCH + 0x0A8 + 0x10)
#define BNX2I_TPAT_IDLE_COUNTER		(BNX2_TPAT_SCRATCH + 0x048 + 0x810)
#define BNX2I_RXP_IDLE_COUNTER		(BNX2_RXP_SCRATCH + 0x0E8 + 0x10)
#define BNX2I_TXP_IDLE_COUNTER		(BNX2_TXP_SCRATCH + 0x048 + 0x10)

#define BNX2I_TXP_TDMA_IDLE_COUNTER	(BNX2I_TXP_IDLE_COUNTER + 0x8)
#define BNX2I_TXP_CTX_IDLE_COUNTER	(BNX2I_TXP_IDLE_COUNTER + 0x10)
#define BNX2I_TXP_HDRQ_IDLE_COUNTER	(BNX2I_TXP_IDLE_COUNTER + 0x18)

#define CTX_READ_TIMEOUT		2 * HZ

#define BNX2I_CP_CTX_ADDR_HI			(BNX2_CP_SCRATCH + 0x68)
#define BNX2I_CP_CTX_ADDR_LO			(BNX2_CP_SCRATCH + 0x6C)
#define BNX2I_CP_CTX_ID				(BNX2_CP_SCRATCH + 0x70)



static u32
bnx2i_reg_rd_ind(struct bnx2i_hba *hba, u32 offset)
{
        u32 val;

        spin_lock_bh(&hba->lock);
        REG_WR(hba, BNX2_PCICFG_REG_WINDOW_ADDRESS, offset);
        val = REG_RD(hba, BNX2_PCICFG_REG_WINDOW);
        spin_unlock_bh(&hba->lock);
        return val;
}

static void
bnx2i_reg_wr_ind(struct bnx2i_hba *hba, u32 offset, u32 val)
{
        spin_lock_bh(&hba->lock);
        REG_WR(hba, BNX2_PCICFG_REG_WINDOW_ADDRESS, offset);
        REG_WR(hba, BNX2_PCICFG_REG_WINDOW, val);
        spin_unlock_bh(&hba->lock);
}

static void bnx2i_read_idle_counter(struct bnx2i_hba *hba, u32 offset,
				    u64 *counter)
{
	u32 cntr32_lsb;
	u32 cntr32_msb;
	cntr32_lsb = bnx2i_reg_rd_ind(hba, offset);
	cntr32_msb = bnx2i_reg_rd_ind(hba, (offset + 4));
	*counter = ((u64) cntr32_lsb << 32) | (u64) cntr32_msb;
}

void bnx2i_read_mips_idle_counters(struct bnx2i_hba *hba)
{
	bnx2i_read_idle_counter(hba, BNX2I_CP_IDLE_COUNTER,
			  	&hba->mips_idle.cp_idle_count);
	bnx2i_read_idle_counter(hba, BNX2I_TXP_IDLE_COUNTER,
			  	&hba->mips_idle.txp_idle_count);
	bnx2i_read_idle_counter(hba, BNX2I_TXP_TDMA_IDLE_COUNTER,
			  	&hba->mips_idle.txp_tdma_count);
	bnx2i_read_idle_counter(hba, BNX2I_TXP_CTX_IDLE_COUNTER,
			  	&hba->mips_idle.txp_ctx_count);
	bnx2i_read_idle_counter(hba, BNX2I_TXP_HDRQ_IDLE_COUNTER,
			  	&hba->mips_idle.txp_hdrq_count);
	bnx2i_read_idle_counter(hba, BNX2I_TPAT_IDLE_COUNTER,
			  	&hba->mips_idle.tpat_idle_count);
	bnx2i_read_idle_counter(hba, BNX2I_RXP_IDLE_COUNTER,
			  	&hba->mips_idle.rxp_idle_count);
	bnx2i_read_idle_counter(hba, BNX2I_COM_IDLE_COUNTER,
			  	&hba->mips_idle.com_idle_count);
}


void bnx2i_init_mips_idle_counters(struct bnx2i_hba *hba)
{
	memset(&hba->mips_idle, 0x00, sizeof(struct mips_idle_count));
}


int bnx2i_select_ctx_dump_cid(struct bnx2i_hba *hba, u32 iscsi_cid)
{
	struct bnx2i_conn *conn;
	u32 addr_lo, addr_hi;
	u32 cid, read_cid;
	int ret = 0;
	int i = 2 * HZ;

	if (!hba->regview)
		return -EPERM;

	if (hba->ctx_read_cnt % 4) {
		printk(KERN_INFO "bnx2i - ctx read, previous request was ");
		printk(KERN_INFO "partially consumed, %d\n",
				 hba->ctx_read_cnt % 4);
	}

	conn = bnx2i_get_conn_from_id(hba, iscsi_cid);
	if (!conn) {
		printk(KERN_INFO "bnx2i - ctx read, iscsi_cid #%d "
				 "not active\n", iscsi_cid);
		return -EINVAL;
	}

	hba->ctx_read_cnt = 0;
	cid = conn->ep->ep_cid;
	addr_lo = (u64) hba->ctx_dma_hndl & 0xffffffff;
	addr_hi = (u64) hba->ctx_dma_hndl >> 32;

	bnx2i_reg_wr_ind(hba, BNX2I_CP_CTX_ADDR_HI, addr_hi);
	bnx2i_reg_wr_ind(hba, BNX2I_CP_CTX_ADDR_LO, addr_lo);
	bnx2i_reg_wr_ind(hba, BNX2I_CP_CTX_ID, cid);
	msleep(10);
	read_cid = bnx2i_reg_rd_ind(hba, BNX2I_CP_CTX_ID);
	while (!read_cid && i--) {
		msleep(1000 / HZ);
		read_cid = bnx2i_reg_rd_ind(hba, BNX2I_CP_CTX_ID);
	}

	if (read_cid) {
		ret = -1;
		printk(KERN_INFO "bnx2i - error in context read\n");
	}

	return ret;
}


int bnx2i_list_iscsi_cid(struct bnx2i_hba *hba, u32 *iscsi_cid_arr,
			 u32 *cid_arr)
{
	int i, idx;
	struct bnx2i_conn *conn;

	for (i = 0, idx = 0; i < hba->cid_que.cid_q_max_idx; i++) {
		if (hba->cid_que.conn_cid_tbl[i]) {
			conn = hba->cid_que.conn_cid_tbl[i];
			iscsi_cid_arr[idx] = i;
			if (conn->ep)
				cid_arr[idx] = conn->ep->ep_cid;
			idx++;
		}
	}
	return idx;
}


/*
 * mark iscsi conn encountered error, this 'iscsi_cid' will be notified
 * to 'ictx' apps next time it polls
 */
void bnx2i_setup_ictx_dump(struct bnx2i_hba *hba,
			   struct bnx2i_conn *conn)
{
	if (hba->ictx_poll_mode && !hba->ictx_poll_cid) {
		hba->ictx_poll_cid = conn->ep->ep_cid;
		hba->ictx_poll_iscsi_cid = conn->ep->ep_iscsi_cid;
	} else {
		printk(KERN_ALERT "conn_err - cnxt dump module is either");
		printk(KERN_ALERT "busy or not in poll mode\n");
	}
}


static ssize_t bnx2i_show_mips_status(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);
	ssize_t len;

	if (!hba->regview)
		return 0;

	bnx2i_read_mips_idle_counters(hba);

	len = sprintf(buf, "%d\n%lu\n%d\n%llu\n%llu\n"
			   "%llu\n%llu\n%llu\n%llu\n%llu\n%llu\n",
			   BNX2I_SYSFS_VERSION, jiffies, HZ,
			   hba->mips_idle.cp_idle_count,
			   hba->mips_idle.txp_idle_count,
			   hba->mips_idle.txp_tdma_count,
			   hba->mips_idle.txp_ctx_count,
			   hba->mips_idle.txp_hdrq_count,
			   hba->mips_idle.tpat_idle_count,
			   hba->mips_idle.rxp_idle_count,
			   hba->mips_idle.com_idle_count);
	return len;
}

static ssize_t bnx2i_show_net_if_name(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	return sprintf(buf, "%s\n", hba->netdev->name);
}

static ssize_t bnx2i_show_sq_info(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	return sprintf(buf, "0x%x\n", hba->max_sqes);
}

static ssize_t bnx2i_set_sq_info(struct class_device *cdev,
				 const char *buf, size_t count)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);
	u32 val;
	int max_sq_size;

	if (test_bit(BNX2I_NX2_DEV_57710, &hba->cnic_dev_type))
		max_sq_size = BNX2I_5770X_SQ_WQES_MAX;
	else
		max_sq_size = BNX2I_570X_SQ_WQES_MAX;

	if (sscanf(buf, " 0x%x ", &val) > 0) {
		if ((val >= BNX2I_SQ_WQES_MIN) && (val <= max_sq_size ))
			hba->max_sqes = val;
	}
	return count;
}

static ssize_t bnx2i_show_cq_info(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	return sprintf(buf, "0x%x\n", hba->max_cqes);
}

static ssize_t bnx2i_set_cq_info(struct class_device *cdev,
				 const char *buf, size_t count)
{
	u32 val;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	if (sscanf(buf, " 0x%x ", &val) > 0) {
		if ((val >= BNX2I_CQ_WQES_MIN) &&
		    (val <= BNX2I_CQ_WQES_MAX)) {
			hba->max_cqes = val;
		}
	}
	return count;
}

static ssize_t bnx2i_show_rq_info(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	return sprintf(buf, "0x%x\n", hba->max_rqes);
}

static ssize_t bnx2i_set_rq_info(struct class_device *cdev, const char *buf,
							size_t count)
{
	u32 val;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	if (sscanf(buf, " 0x%x ", &val) > 0) {
		if ((val >= BNX2I_RQ_WQES_MIN) &&
		    (val <= BNX2I_RQ_WQES_MAX)) {
			hba->max_rqes = val;
		}
	}
	return count;
}


static ssize_t bnx2i_show_ccell_info(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	return sprintf(buf, "0x%x\n", hba->num_ccell);
}

static ssize_t bnx2i_set_ccell_info(struct class_device *cdev,
				    const char *buf, size_t count)
{
	u32 val;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	if (sscanf(buf, " 0x%x ", &val) > 0) {
		if ((val >= BNX2I_CCELLS_MIN) &&
		    (val <= BNX2I_CCELLS_MAX)) {
			hba->num_ccell = val;
		}
	}
	return count;
}


static ssize_t bnx2i_read_pci_trigger_reg(struct class_device *cdev,
					  char *buf)
{
	u32 reg_val;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	if (!hba->regview)
		return 0;

#define PCI_EVENT_TRIGGER_REG	0xCAC	/* DMA WCHAN STAT10 REG */
	reg_val = readl(hba->regview + PCI_EVENT_TRIGGER_REG);
	return sprintf(buf, "0x%x\n", reg_val);
}


static ssize_t bnx2i_get_iscsi_cntx_dump(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);
	unsigned int *ptr = (unsigned int *) hba->ctx_addr;
	unsigned int *dst_ptr = (unsigned int *) buf;
	int unit_sz = sizeof(unsigned int);
#define SYSFS_BUF_SIZE			4096
#define NUM_SYSFS_BUFS_PER_CTX		4

	if ((hba->ctx_read_cnt == NUM_SYSFS_BUFS_PER_CTX) || !ptr)
		return 0;

	ptr += (((hba->ctx_read_cnt % NUM_SYSFS_BUFS_PER_CTX) *
		 SYSFS_BUF_SIZE) / unit_sz);
	hba->ctx_read_cnt++;
	memcpy(dst_ptr, ptr, SYSFS_BUF_SIZE);

	return SYSFS_BUF_SIZE;
}

static ssize_t bnx2i_select_iscsi_cntx_dump(struct class_device *cdev,
					    const char *buf, size_t count)
{
	u32 iscsi_cid;
	int ret = 0;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	if (sscanf(buf, " 0x%x ", &iscsi_cid) > 0) {
		ret = bnx2i_select_ctx_dump_cid(hba, iscsi_cid);
	}
	if (!ret)
		ret = count;
	return ret;
}

static ssize_t bnx2i_get_active_iscsi_cid_list(struct class_device *cdev,
					       char *buf)
{
	u32 active_iscsi_cid[32];
	u32 active_cid[32];
	int num_cid = 0;
	ssize_t total_len = 0;
	char *cur_ptr = buf;
	int i;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);
	u32 num_ccell = hba->ctx_ccell_tasks & 0xFFFF;
	u32 num_tasks_per_conn = hba->ctx_ccell_tasks >> 16;

	if (!hba->ictx_poll_mode) {
		num_cid = bnx2i_list_iscsi_cid(hba, active_iscsi_cid,
					       active_cid);
	}
	total_len += sprintf(cur_ptr, "0x%x\n", BNX2I_SYSFS_VERSION);
	cur_ptr = buf + total_len;
	total_len += sprintf(cur_ptr, "0x%x\n", num_ccell);
	cur_ptr = buf + total_len;
	total_len += sprintf(cur_ptr, "0x%x\n", num_tasks_per_conn);
	if (hba->ictx_poll_mode) {
		if (hba->ictx_poll_cid) {
			cur_ptr = buf + total_len;
			total_len += sprintf(cur_ptr, "0x%x, 0x%x\n",
					     hba->ictx_poll_iscsi_cid,
					     hba->ictx_poll_cid);
			hba->ictx_poll_cid = hba->ictx_poll_iscsi_cid = 0;
		}
	} else {
		for (i = 0; i < num_cid; i++) {
			cur_ptr = buf + total_len;
			total_len += sprintf(cur_ptr, "0x%x, 0x%x\n",
					     active_iscsi_cid[i],
					     active_cid[i]);
		}
	}
	return total_len;
}


static ssize_t bnx2i_set_iscsi_cid_err_poll_mode(struct class_device *cdev,
						 const char *buf, size_t count)
{
	u32 poll_mode;
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);

	if (sscanf(buf, "0x%x", &poll_mode) > 0) {
		if (poll_mode)
			hba->ictx_poll_mode = 1;
		else
			hba->ictx_poll_mode = 0;
	}
	return count;
}


static ssize_t bnx2i_get_qp_shmem_dump(struct class_device *cdev, char *buf)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);
	int resi_len = hba->sq_cq_size -
		       (hba->sq_cq_rdp - hba->sq_cq_dump);

	if (!hba->sq_cq_dump || !hba->sq_cq_rdp) {
		return -EINVAL;
	} else if ((hba->sq_cq_dump + hba->sq_cq_size) ==
		   hba->sq_cq_rdp) {
		kfree(hba->sq_cq_dump);
		hba->sq_cq_dump = hba->sq_cq_rdp = NULL;
		return 0;
	}

	if (resi_len > SYSFS_BUF_SIZE) {
		resi_len = SYSFS_BUF_SIZE;
	}
	memcpy(buf, hba->sq_cq_rdp, resi_len);
	hba->sq_cq_rdp += resi_len;

	return resi_len;
}




static void bnx2i_dup_cq_mem(struct bnx2i_hba *hba,
			     struct bnx2i_conn *conn, int count)
{
	struct cqe *cqe_s;
	struct cqe *cqe_d;
	int total_cnt = count;

	if (conn->ep->qp.cq_cons_qe == conn->ep->qp.cq_virt)
		cqe_s = conn->ep->qp.cq_last_qe;
	else
		cqe_s = conn->ep->qp.cq_cons_qe - 1;
	cqe_d = (struct cqe *)hba->sq_cq_rdp;
	while (count--) {
		memcpy(cqe_d, cqe_s, sizeof(struct cqe));
		if (cqe_s == conn->ep->qp.cq_virt)
			cqe_s = conn->ep->qp.cq_last_qe;
		else
			cqe_s--;
		cqe_d++;
		if ((cqe_d - (struct cqe *)hba->sq_cq_rdp) > total_cnt) {
			printk(KERN_ALERT "bnx2i - SQ Dump: mem overflow\n");
			break;
		}
	}
}


static int bnx2i_init_cq_dump(struct bnx2i_hba *hba, u32 iscsi_cid,
			      u32 count)
{
	struct bnx2i_conn *conn;
	int cq_size;

	conn = bnx2i_get_conn_from_id(hba, iscsi_cid);

	if (!conn) {
		printk(KERN_ALERT "CQ dump: cid #%x not valid\n",
				  iscsi_cid);
		return -EPERM;
	}

	if (hba->sq_cq_dump)
		return -EPERM;

	cq_size = (conn->ep->qp.cq_last_qe - conn->ep->qp.cq_first_qe) + 1;

	if (!count || (count > cq_size))
		count = cq_size;

	hba->sq_cq_size = count * sizeof(struct cqe);

	if (!hba->sq_cq_size)
		return -EINVAL;

	hba->sq_cq_dump = kmalloc(hba->sq_cq_size, GFP_KERNEL);
	if (!hba->sq_cq_dump)
		return -ENOMEM;
	hba->sq_cq_rdp = hba->sq_cq_dump;

	bnx2i_dup_cq_mem(hba, conn, count);
	return 0;
}



static void bnx2i_dup_sq_mem(struct bnx2i_hba *hba,
			     struct bnx2i_conn *conn, int count)
{
	struct sqe *sqe_s;
	struct sqe *sqe_d;
	int total_cnt = count;

	if (conn->ep->qp.sq_prod_qe == conn->ep->qp.sq_virt)
		sqe_s = conn->ep->qp.sq_last_qe;
	else
		sqe_s = conn->ep->qp.sq_prod_qe - 1;
	sqe_d = (struct sqe *)hba->sq_cq_rdp;
	while (count--) {
		memcpy(sqe_d, sqe_s, sizeof(struct sqe));
		if (sqe_s == conn->ep->qp.sq_virt) {
			sqe_s = conn->ep->qp.sq_last_qe;
		} else {
			sqe_s--;
		}
		sqe_d++;
		if ((sqe_d - (struct sqe *) hba->sq_cq_rdp) >
		    total_cnt) {
			printk(KERN_ALERT "bnx2i - SQ Dump: mem overflow\n");
			break;
		}
	}
}


static int bnx2i_init_sq_dump(struct bnx2i_hba *hba,
			      u32 iscsi_cid, u32 count)
{
	struct bnx2i_conn *conn;
	int sq_size;

	conn = bnx2i_get_conn_from_id(hba, iscsi_cid);

	if (!conn) {
		printk(KERN_ALERT "SQ dump: cid #%x not valid\n",
				  iscsi_cid);
		return -EINVAL;
	}

	if (hba->sq_cq_dump)
		return -EINVAL;

	sq_size = (conn->ep->qp.sq_last_qe - conn->ep->qp.sq_first_qe) + 1;

	if (!count || (count > sq_size))
		count = sq_size;

	hba->sq_cq_size = count * sizeof(struct sqe);

	if (!hba->sq_cq_size)
		return -EINVAL;

	hba->sq_cq_dump = kmalloc(hba->sq_cq_size, GFP_KERNEL);
	if (!hba->sq_cq_dump)
		return -ENOMEM;
	hba->sq_cq_rdp = hba->sq_cq_dump;

	bnx2i_dup_sq_mem(hba, conn, count);
	return 0;
}

static ssize_t bnx2i_setup_qp_shmem_dump(struct class_device *cdev,
					 const char *buf, size_t count)
{
	struct bnx2i_hba *hba =
		container_of(cdev, struct bnx2i_hba, class_dev);
	u32 iscsi_cid;
	char queue[32];
	ssize_t ret = count;
	u32 num_count;


	if (sscanf(buf, "%c%c,%d,%d", &queue[0], &queue[1],
		   &iscsi_cid, &num_count) > 0) {
		if (!strncmp(queue, "SQ", 2)) {
			ret = bnx2i_init_sq_dump(hba, iscsi_cid, num_count);
		} else if (!strncmp(queue, "CQ", 2)) {
			ret = bnx2i_init_cq_dump(hba, iscsi_cid, num_count);
		} else {
			ret = -EINVAL;
		}
	}
	return ret;
}


static ssize_t bnx2i_read_tcp_portd_options(struct class_device *cdev,
					    char *buf)
{
	extern struct tcp_port_mngt bnx2i_tcp_port_tbl;
	return sprintf(buf, "0x%x\n", bnx2i_tcp_port_tbl.num_required);
}

static ssize_t bnx2i_write_tcp_portd_results(struct class_device *cdev,
					     const char *buf, size_t count)
{
	extern struct tcp_port_mngt bnx2i_tcp_port_tbl;
	u32 tcp_port, bind_stat;

	if (!bnx2i_tcp_port_tbl.free_q)
		return count;

	if (sscanf(buf, "%d,%d", &tcp_port, &bind_stat) > 0) {
		if (bind_stat && tcp_port) {
			bnx2i_tcp_port_new_entry(tcp_port);
		}
	}
	return count;
}


static CLASS_DEVICE_ATTR (mips_info, S_IRUGO,
			 bnx2i_show_mips_status, NULL);
static CLASS_DEVICE_ATTR (net_if_name, S_IRUGO,
			 bnx2i_show_net_if_name, NULL);
static CLASS_DEVICE_ATTR (sq_size, S_IRUGO | S_IWUSR,
			 bnx2i_show_sq_info, bnx2i_set_sq_info);
static CLASS_DEVICE_ATTR (cq_size, S_IRUGO | S_IWUSR,
			 bnx2i_show_cq_info, bnx2i_set_cq_info);
static CLASS_DEVICE_ATTR (rq_size, S_IRUGO | S_IWUSR,
			 bnx2i_show_rq_info, bnx2i_set_rq_info);
static CLASS_DEVICE_ATTR (num_ccell, S_IRUGO | S_IWUSR,
			 bnx2i_show_ccell_info, bnx2i_set_ccell_info);
static CLASS_DEVICE_ATTR (pci_trigger, S_IRUGO,
			 bnx2i_read_pci_trigger_reg, NULL);
static CLASS_DEVICE_ATTR (ctx_dump, S_IRUGO | S_IWUSR,
			 bnx2i_get_iscsi_cntx_dump,
			 bnx2i_select_iscsi_cntx_dump);
static CLASS_DEVICE_ATTR (cid_list, S_IRUGO | S_IWUSR,
			 bnx2i_get_active_iscsi_cid_list,
			 bnx2i_set_iscsi_cid_err_poll_mode);
static CLASS_DEVICE_ATTR (qp_shmem_dump, S_IRUGO | S_IWUSR,
			 bnx2i_get_qp_shmem_dump,
			 bnx2i_setup_qp_shmem_dump);
static CLASS_DEVICE_ATTR (port_bind, S_IRUGO | S_IWUSR,
			 bnx2i_read_tcp_portd_options,
			 bnx2i_write_tcp_portd_results);


static struct class_device_attribute *bnx2i_class_attributes[] = {
	&class_device_attr_mips_info,
	&class_device_attr_net_if_name,
	&class_device_attr_sq_size,
	&class_device_attr_cq_size,
	&class_device_attr_rq_size,
	&class_device_attr_num_ccell,
	&class_device_attr_pci_trigger,
	&class_device_attr_ctx_dump,
	&class_device_attr_cid_list,
	&class_device_attr_qp_shmem_dump,
};

static struct class_device_attribute *tcp_port_class_attributes[] = {
	&class_device_attr_port_bind
};

static void bnx2i_sysfs_release(struct class_device *class_dev)
{
}

struct class_device port_class_dev;


static struct class bnx2i_class = {
	.name	= "bnx2i",
	.release = bnx2i_sysfs_release,
};



static int bnx2i_register_port_class_dev(struct class_device *class_dev)
{
	char dev_name[BUS_ID_SIZE];
	int ret;
	int i;

	class_dev->class = &bnx2i_class;
	class_dev->class_data = class_dev;
	snprintf(dev_name, BUS_ID_SIZE, "%s", "tcp_portd");
	strlcpy(class_dev->class_id, dev_name, BUS_ID_SIZE);

	ret = class_device_register(class_dev);
	if (ret)
		goto err;

	for (i = 0; i < ARRAY_SIZE(tcp_port_class_attributes); ++i) {
		ret = class_device_create_file(class_dev,
					       tcp_port_class_attributes[i]);
		if (ret)
			goto err_unregister;
	}

	return 0;

err_unregister:
	class_device_unregister(class_dev);
err:
	return ret;
}


int bnx2i_register_sysfs(struct bnx2i_hba *hba)
{
	struct class_device *class_dev = &hba->class_dev;
	char dev_name[BUS_ID_SIZE];
	struct ethtool_drvinfo drv_info;
	u32 bus_no;
	u32 dev_no;
	u32 func_no;
	u32 extra;
	int ret;
	int i;

	if (hba->cnic && hba->cnic->netdev) {
		hba->cnic->netdev->ethtool_ops->get_drvinfo(hba->cnic->netdev,
							    &drv_info);
		sscanf(drv_info.bus_info, "%x:%x:%x.%d",
		       &extra, &bus_no, &dev_no, &func_no);
	}
	class_dev->class = &bnx2i_class;
	class_dev->class_data = hba;
	snprintf(dev_name, BUS_ID_SIZE, "%.2x:%.2x.%.1x",
			 bus_no, dev_no, func_no);
	strlcpy(class_dev->class_id, dev_name, BUS_ID_SIZE);

	ret = class_device_register(class_dev);
	if (ret)
		goto err;

	for (i = 0; i < ARRAY_SIZE(bnx2i_class_attributes); ++i) {
		ret = class_device_create_file(class_dev,
					       bnx2i_class_attributes[i]);
		if (ret)
			goto err_unregister;
	}

	return 0;

err_unregister:
	class_device_unregister(class_dev);
err:
	return ret;
}

void bnx2i_unregister_sysfs(struct bnx2i_hba *hba)
{
	class_device_unregister(&hba->class_dev);
}

int bnx2i_sysfs_setup(void)
{
	int ret;
	ret = class_register(&bnx2i_class);

	bnx2i_register_port_class_dev(&port_class_dev);
	return ret;
}

void bnx2i_sysfs_cleanup(void)
{
	class_device_unregister(&port_class_dev);
	class_unregister(&bnx2i_class);
}

#endif
